from __future__ import annotations

import logging
import typing
import xml.etree.ElementTree as ET
import contextlib
from pathlib import Path
import os
import abc
import functools
import warnings

from mpi4py import MPI
import numpy as np
import numpy.typing as npt
import dolfinx

logger = logging.getLogger(__name__)


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        msg = (
            "Call to deprecated function {}.".format(func.__name__),
            "Please use the class scifem.xdmf.XDMFFile instead.",
        )
        logger.warning(msg)
        warnings.warn(
            msg,
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


class XDMFData(typing.Protocol):
    @property
    def points(self) -> npt.NDArray[np.float64]: ...

    @property
    def bs(self) -> int: ...

    @property
    def num_dofs_global(self) -> int: ...


@deprecated
def write_xdmf(
    functions: typing.Sequence[dolfinx.fem.Function],
    filename: os.PathLike,
    h5name: Path,
    xdmfdata: XDMFData,
) -> None:
    """Write the XDMF file for the point cloud.

    Args:
        functions: List of functions to write to the point cloud.
        filename: The name of the file to write the XDMF to.
        h5name: The name of the HDF5 file.
        xdmfdata: The XDMF data.

    Note:
        This function does not check the validity of the input data, i.e.
        all functions in ``functions`` are assumed to share the same function space,
        including ``block_size``.
    """

    xdmf = ET.Element("XDMF")
    xdmf.attrib["Version"] = "3.0"
    xdmf.attrib["xmlns:xi"] = "http://www.w3.org/2001/XInclude"
    domain = ET.SubElement(xdmf, "Domain")
    grid = ET.SubElement(domain, "Grid")
    grid.attrib["GridType"] = "Uniform"
    grid.attrib["Name"] = "Point Cloud"
    topology = ET.SubElement(grid, "Topology")
    topology.attrib["NumberOfElements"] = str(xdmfdata.num_dofs_global)
    topology.attrib["TopologyType"] = "PolyVertex"
    topology.attrib["NodesPerElement"] = "1"
    top_data = ET.SubElement(topology, "DataItem")
    top_data.attrib["Dimensions"] = f"{xdmfdata.num_dofs_global} {1}"
    top_data.attrib["Format"] = "HDF"
    top_data.text = f"{h5name.name}:/Step0/Cells"
    geometry = ET.SubElement(grid, "Geometry")
    geometry.attrib["GeometryType"] = "XYZ"
    for u in functions:
        it0 = ET.SubElement(geometry, "DataItem")
        it0.attrib["Dimensions"] = f"{xdmfdata.num_dofs_global} {xdmfdata.points.shape[1]}"
        it0.attrib["Format"] = "HDF"
        it0.text = f"{h5name.name}:/Step0/Points"
        attrib = ET.SubElement(grid, "Attribute")
        attrib.attrib["Name"] = u.name
        out_bs = xdmfdata.bs
        if out_bs == 1:
            attrib.attrib["AttributeType"] = "Scalar"
        else:
            if out_bs == 2:
                # Pad to 3D
                out_bs = 3
            attrib.attrib["AttributeType"] = "Vector"
        attrib.attrib["Center"] = "Node"
        it1 = ET.SubElement(attrib, "DataItem")
        it1.attrib["Dimensions"] = f"{xdmfdata.num_dofs_global} {out_bs}"
        it1.attrib["Format"] = "HDF"
        it1.text = f"{h5name.name}:/Step0/Values_{u.name}"
    text = [
        '<?xml version="1.0"?>\n<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n',
        ET.tostring(xdmf, encoding="unicode"),
    ]
    Path(filename).with_suffix(".xdmf").write_text("".join(text))


class FunctionSpaceData(typing.NamedTuple):
    """Data class for function space information."""

    points: npt.NDArray[np.float64]
    bs: int
    num_dofs_global: int
    num_dofs_local: int
    local_range: npt.NDArray[np.int64]
    comm: MPI.Intracomm

    @property
    def points_out(self) -> npt.NDArray[np.float64]:
        """Pad points to be 3D"""
        return np.ascontiguousarray(self.points[: self.num_dofs_local, :])


@contextlib.contextmanager
def h5pyfile(h5name, filemode="r", force_serial: bool = False, comm=None):
    """Context manager for opening an HDF5 file with h5py.

    Args:
        h5name: The name of the HDF5 file.
        filemode: The file mode.
        force_serial: Force serial access to the file.
        comm: The MPI communicator

    """
    import h5py

    if comm is None:
        comm = MPI.COMM_WORLD

    if h5py.h5.get_config().mpi and comm.size > 1 and not force_serial:
        h5file = h5py.File(h5name, filemode, driver="mpio", comm=comm)
    else:
        if comm.size > 1 and not force_serial:
            raise ValueError(
                f"h5py is not installed with MPI support, while using {comm.size} processes.",
                "If you really want to do this, turn on the `force_serial` flag.",
            )
        h5file = h5py.File(h5name, filemode)
    yield h5file
    h5file.close()


@deprecated
def write_hdf5_h5py(
    functions: typing.Sequence[dolfinx.fem.Function],
    h5name: Path,
    data: FunctionSpaceData | None = None,
) -> None:
    """Write the point cloud to an HDF5 file using h5py.

    Args:
        functions: The functions to write to the point cloud.
        h5name: The name of the file to write the point cloud to.
        data: The function space data.

    Note:
        All input ``functions`` has to share the same function space.
    """
    if len(functions) == 0:
        return

    if data is None:
        data = check_function_space(functions)
    if data is None:
        warnings.warn("No functions to write to point cloud")
        return

    with h5pyfile(h5name=h5name, filemode="w", comm=data.comm) as h5file:
        step = h5file.create_group(np.bytes_("Step0"))
        points = step.create_dataset(
            "Points", (data.num_dofs_global, data.points.shape[1]), dtype=data.points.dtype
        )
        points[data.local_range[0] : data.local_range[1], :] = data.points_out
        cells = step.create_dataset("Cells", (data.num_dofs_global,), dtype=np.int64)
        cells[data.local_range[0] : data.local_range[1]] = np.arange(
            data.local_range[0], data.local_range[1], dtype=np.int64
        )
        for u in functions:
            # Pad array to 3D if vector space with 2 components
            array = np.zeros(
                (data.num_dofs_local, data.bs if data.bs != 2 else 3), dtype=u.x.array.dtype
            )
            array[:, : data.bs] = u.x.array[: data.num_dofs_local * data.bs].reshape(-1, data.bs)
            dset = step.create_dataset(
                f"Values_{u.name}", (data.num_dofs_global, array.shape[1]), dtype=array.dtype
            )
            dset[data.local_range[0] : data.local_range[1], :] = array


@deprecated
def write_hdf5_adios(
    functions: typing.Sequence[dolfinx.fem.Function],
    h5name: Path,
    data: FunctionSpaceData | None = None,
) -> None:
    """Write the point cloud to an HDF5 file using ADIOS2.

    Args:
        functions: The functions to write to the point cloud.
        h5name: The name of the file to write the point cloud to.
        data: The function space data.

    Note:
        All input ``functions`` has to share the same function space.
    """
    import adios2

    def resolve_adios_scope(adios2):
        return adios2 if not hasattr(adios2, "bindings") else adios2.bindings

    adios2 = resolve_adios_scope(adios2)

    if len(functions) == 0:
        return

    if data is None:
        data = check_function_space(functions)
    if data is None:
        warnings.warn("No functions to write to point cloud")
        return

    # Create ADIOS2 reader
    adios = adios2.ADIOS(data.comm)
    io = adios.DeclareIO("Point cloud writer")
    io.SetEngine("HDF5")
    outfile = io.Open(h5name.as_posix(), adios2.Mode.Write)

    pointvar = io.DefineVariable(
        "Points",
        data.points_out,
        shape=[data.num_dofs_global, data.points.shape[1]],
        start=[data.local_range[0], 0],
        count=[data.num_dofs_local, data.points.shape[1]],
    )
    outfile.Put(pointvar, data.points_out)
    cells = np.arange(data.local_range[0], data.local_range[1], dtype=np.int64)
    cellvar = io.DefineVariable(
        "Cells",
        cells,
        shape=[data.num_dofs_global],
        start=[data.local_range[0]],
        count=[data.num_dofs_local],
    )
    outfile.Put(cellvar, cells)

    for u in functions:
        array = np.zeros(
            (data.num_dofs_local, data.bs if data.bs != 2 else 3), dtype=u.x.array.dtype
        )
        array[:, : data.bs] = u.x.array[: data.num_dofs_local * data.bs].reshape(-1, data.bs)

        valuevar = io.DefineVariable(
            f"Values_{u.name}",
            array,
            shape=[data.num_dofs_global, array.shape[1]],
            start=[data.local_range[0], 0],
            count=[data.num_dofs_local, array.shape[1]],
        )
        outfile.Put(valuevar, array)
    outfile.PerformPuts()
    outfile.Close()
    assert adios.RemoveIO("Point cloud writer")


def create_function_space_data(V: dolfinx.fem.FunctionSpace) -> FunctionSpaceData:
    """Create function space data from a function space.

    Args:
        V: The function space.

    Returns:
        The function space data.

    """
    points = V.tabulate_dof_coordinates()
    bs = V.dofmap.index_map_bs
    index_map = V.dofmap.index_map
    comm = V.mesh.comm

    return FunctionSpaceData(
        points=points,
        bs=bs,
        num_dofs_global=index_map.size_global,
        num_dofs_local=index_map.size_local,
        local_range=np.array(index_map.local_range, dtype=np.int64),
        comm=comm,
    )


def check_function_space(
    functions: typing.Sequence[dolfinx.fem.Function],
) -> FunctionSpaceData | None:
    """Check that all functions are in the same function space,
    and return the function space data.

    Args:
        functions: The functions to check.

    Returns:
        The function space data if all functions are in the same function space, otherwise None.

    """
    if len(functions) == 0:
        return None

    # Check that all functions are in the same function space
    u = functions[0]
    for v in functions[1:]:
        if u.function_space != v.function_space:
            raise ValueError("All functions must be in the same function space.")

    return create_function_space_data(u.function_space)


@deprecated
def create_pointcloud(
    filename: os.PathLike, functions: typing.Sequence[dolfinx.fem.Function]
) -> None:
    """Create a point cloud from a list of functions to be visualized in Paraview.
    The point cloud is written to a file in XDMF format.

    Args:
        filename: The name of the file to write the point cloud to.
        functions: The functions to write to the point cloud.

    Note:
        This is useful for visualizing functions in quadrature spaces.

    Note:
        Any function space that can call `tabulate_dof_coordinates` can be used.

    Note:
        ADIOS2 is the preferred backend for writing HDF5 files, and will be used if available.
        If ADIOS2 is not available, `h5py` will be used.
    """
    data = check_function_space(functions)
    if data is None:
        warnings.warn("No functions to write to point cloud")
        return

    h5name = Path(filename).with_suffix(".h5")

    # Write XDMF on rank 0
    if data.comm.rank == 0:
        write_xdmf(functions, filename, h5name, data)

    try:
        write_hdf5_adios(functions=functions, h5name=h5name, data=data)
    except (ImportError, TypeError, ValueError):
        warnings.warn("ADIOS2 not available, using h5py", stacklevel=2)
        write_hdf5_h5py(functions=functions, h5name=h5name, data=data)


class BaseXDMFFile(abc.ABC):
    filename: os.PathLike
    filemode: typing.Literal["r", "a", "w"]
    backend: typing.Literal["h5py", "adios2"]
    _data: FunctionSpaceData

    @property
    @abc.abstractmethod
    def data_names(self) -> list[str]:
        """The names of the data."""

    @property
    @abc.abstractmethod
    def data_arrays(self) -> list[npt.NDArray[np.floating]]:
        """The data arrays."""
        ...

    def __post_init__(self) -> None:
        self.h5name = Path(self.filename).with_suffix(".h5")
        self.xdmfname = Path(self.filename).with_suffix(".xdmf")

        if self.filemode == "r":
            if not self.h5name.exists():
                raise FileNotFoundError(f"{self.h5name} does not exist")
            if not self.xdmfname.exists():
                raise FileNotFoundError(f"{self.xdmfname} does not exist")
        elif self.filemode == "w":
            # Overwrite existing files so make sure they don't exist
            self.h5name.unlink(missing_ok=True)
            self.xdmfname.unlink(missing_ok=True)

        self._time_values: dict[float, int] = {}

    def _init_backend(self) -> None:
        assert self.backend in [
            "h5py",
            "adios2",
        ], f"Unknown backend {self.backend}, must be 'h5py' or 'adios2'"
        if self.backend == "adios2":
            try:
                self._init_adios()
            except (ImportError, TypeError, ValueError):
                msg = "ADIOS2 not available, using h5py"
                warnings.warn(msg)
                logging.warning(msg)
                self.backend = "h5py"
        if self.backend == "h5py":
            self._init_h5py()

    def _init_h5py(self) -> None:
        logger.debug("Initializing h5py")
        self._outfile = h5pyfile(
            h5name=self.h5name, filemode=self.filemode, comm=self._data.comm
        ).__enter__()
        self._step = self._outfile.create_group(np.bytes_("Step0"))
        points = self._step.create_dataset(
            "Points",
            (self._data.num_dofs_global, self._data.points.shape[1]),
            dtype=self._data.points.dtype,
        )
        points[self._data.local_range[0] : self._data.local_range[1], :] = self._data.points_out
        cells = self._step.create_dataset("Cells", (self._data.num_dofs_global,), dtype=np.int64)
        cells[self._data.local_range[0] : self._data.local_range[1]] = np.arange(
            self._data.local_range[0], self._data.local_range[1], dtype=np.int64
        )

    def _write_h5py(self, index: int) -> None:
        logger.debug(f"Writing h5py at time {index}")
        for data_name, data_array in zip(self.data_names, self.data_arrays):
            # Pad array to 3D if vector space with 2 components
            array = np.zeros(
                (self._data.num_dofs_local, self._data.bs if self._data.bs != 2 else 3),
                dtype=data_array.dtype,
            )
            array[:, : self._data.bs] = data_array[
                : self._data.num_dofs_local * self._data.bs
            ].reshape(-1, self._data.bs)
            dset = self._step.create_dataset(
                f"Values_{data_name}_{index}",
                (self._data.num_dofs_global, array.shape[1]),
                dtype=array.dtype,
            )
            dset[self._data.local_range[0] : self._data.local_range[1], :] = array

    def _close_h5py(self) -> None:
        logger.debug("Closing HDF5 file")
        self._outfile.close()

    def _init_adios(self) -> None:
        logger.debug("Initializing ADIOS2")
        import adios2

        def resolve_adios_scope(adios2):
            return adios2 if not hasattr(adios2, "bindings") else adios2.bindings

        adios2 = resolve_adios_scope(adios2)

        # Create ADIOS2 reader
        self._adios = adios2.ADIOS(self._data.comm)
        self._io = self._adios.DeclareIO("Point cloud writer")
        self._io.SetEngine("HDF5")
        self._outfile = self._io.Open(self.h5name.as_posix(), adios2.Mode.Write)
        pointvar = self._io.DefineVariable(
            "Points",
            self._data.points_out,
            shape=[self._data.num_dofs_global, self._data.points.shape[1]],
            start=[self._data.local_range[0], 0],
            count=[self._data.num_dofs_local, self._data.points.shape[1]],
        )
        self._outfile.Put(pointvar, self._data.points_out)
        cells = np.arange(self._data.local_range[0], self._data.local_range[1], dtype=np.int64)
        cellvar = self._io.DefineVariable(
            "Cells",
            cells,
            shape=[self._data.num_dofs_global],
            start=[self._data.local_range[0]],
            count=[self._data.num_dofs_local],
        )
        self._outfile.Put(cellvar, cells)

    def _write_adios(self, index: int) -> None:
        logger.debug(f"Writing adios at time {index}")
        for data_name, data_array in zip(self.data_names, self.data_arrays):
            array = np.zeros(
                (self._data.num_dofs_local, self._data.bs if self._data.bs != 2 else 3),
                dtype=data_array.dtype,
            )
            array[:, : self._data.bs] = data_array[
                : self._data.num_dofs_local * self._data.bs
            ].reshape(-1, self._data.bs)

            valuevar = self._io.DefineVariable(
                f"Values_{data_name}_{index}",
                array,
                shape=[self._data.num_dofs_global, array.shape[1]],
                start=[self._data.local_range[0], 0],
                count=[self._data.num_dofs_local, array.shape[1]],
            )
            self._outfile.Put(valuevar, array)
        self._outfile.PerformPuts()

    def _close_adios(self) -> None:
        logger.debug("Closing ADIOS2 file")
        try:
            self._outfile.Close()
            assert self._adios.RemoveIO("Point cloud writer")
        except ValueError:
            # File is allready closed
            logger.debug("ADIOS2 file already closed")

    def close(self) -> None:
        """Close the XDMF file."""
        logger.debug("Closing XDMF file")
        if self.backend == "adios2":
            self._close_adios()
        elif self.backend == "h5py":
            self._close_h5py()

    def _write_xdmf(self) -> None:
        logger.debug("Writing XDMF file")
        xdmf = ET.Element("Xdmf")
        xdmf.attrib["Version"] = "3.0"
        xdmf.attrib["xmlns:xi"] = "http://www.w3.org/2001/XInclude"
        domain = ET.SubElement(xdmf, "Domain")
        grid = ET.SubElement(domain, "Grid")
        grid.attrib["GridType"] = "Uniform"
        grid.attrib["Name"] = "Point Cloud"
        topology = ET.SubElement(grid, "Topology")
        topology.attrib["NumberOfElements"] = str(self._data.num_dofs_global)
        topology.attrib["TopologyType"] = "PolyVertex"
        topology.attrib["NodesPerElement"] = "1"
        top_data = ET.SubElement(topology, "DataItem")
        top_data.attrib["Dimensions"] = f"{self._data.num_dofs_global} {1}"
        top_data.attrib["Format"] = "HDF"
        top_data.text = f"{self.h5name.name}:/Step0/Cells"
        geometry = ET.SubElement(grid, "Geometry")
        geometry.attrib["GeometryType"] = "XYZ"
        it0 = ET.SubElement(geometry, "DataItem")
        it0.attrib["Dimensions"] = f"{self._data.num_dofs_global} {self._data.points.shape[1]}"
        it0.attrib["Format"] = "HDF"
        it0.text = f"{self.h5name.name}:/Step0/Points"
        for name in self.data_names:
            ugrid_collection = ET.SubElement(domain, "Grid")
            ugrid_collection.attrib["GridType"] = "Collection"
            ugrid_collection.attrib["CollectionType"] = "Temporal"
            ugrid_collection.attrib["Name"] = name

            for step, time_value in enumerate(sorted(self._time_values)):
                ugrid = ET.SubElement(ugrid_collection, "Grid")
                ugrid.attrib["GridType"] = "Uniform"
                ugrid.attrib["Name"] = name
                xp = ET.SubElement(ugrid, "xi:include")
                xp.attrib["xpointer"] = (
                    "xpointer(/Xdmf/Domain/Grid[@GridType='Uniform'][1]/*[self::Topology or self::Geometry])"  # noqa: E501
                )
                time = ET.SubElement(ugrid, "Time")
                time.attrib["Value"] = str(time_value)
                attrib = ET.SubElement(ugrid, "Attribute")
                attrib.attrib["Name"] = name
                out_bs = self._data.bs
                if out_bs == 1:
                    attrib.attrib["AttributeType"] = "Scalar"
                else:
                    if out_bs == 2:
                        # Pad to 3D
                        out_bs = 3
                    attrib.attrib["AttributeType"] = "Vector"
                attrib.attrib["Center"] = "Node"
                it1 = ET.SubElement(attrib, "DataItem")
                it1.attrib["Dimensions"] = f"{self._data.num_dofs_global} {out_bs}"
                it1.attrib["Format"] = "HDF"
                it1.text = f"{self.h5name.name}:/Step0/Values_{name}_{step}"

        ET.indent(xdmf)
        text = [
            '<?xml version="1.0"?>\n<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n',
            ET.tostring(xdmf, encoding="unicode"),
        ]
        self.xdmfname.write_text("".join(text))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def write(self, time: float) -> None:
        """Write the point cloud at a given time.

        Args:
            time: The time value.

        """
        logger.debug(f"Writing time {time}")
        time = float(time)
        if time in self._time_values:
            msg = f"Time {time} already written to file. Skipping."
            logger.warning(msg)
            return

        index = len(self._time_values)
        self._time_values[time] = index

        self._write_xdmf()
        if self.backend == "adios2":
            self._write_adios(index)
        elif self.backend == "h5py":
            self._write_h5py(index)


class XDMFFile(BaseXDMFFile):
    def __init__(
        self,
        filename: os.PathLike,
        functions: typing.Sequence[dolfinx.fem.Function],
        filemode: typing.Literal["r", "a", "w"] = "w",
        backend: typing.Literal["h5py", "adios2"] = "adios2",
    ) -> None:
        self.functions = functions
        self.filename = filename
        self.filemode = filemode
        self.backend = backend
        super().__post_init__()
        if len(self.functions) == 0:
            raise ValueError("No functions to write to point cloud")
        for f in self.functions:
            if not isinstance(f, dolfinx.fem.Function):
                raise ValueError("All functions must be of type dolfinx.fem.Function")

        data = check_function_space(self.functions)
        if data is None:
            raise ValueError("All functions must be in the same function space")
        self._data = data

        self._init_backend()

    @property
    def data_names(self) -> list[str]:
        return [f.name for f in self.functions]

    @property
    def data_arrays(self) -> list[npt.NDArray[np.floating]]:
        return [f.x.array for f in self.functions]


class NumpyXDMFFile(BaseXDMFFile):
    def __init__(
        self,
        filename: os.PathLike,
        arrays: list[npt.NDArray[np.floating]],
        function_space_data: FunctionSpaceData,
        filemode: typing.Literal["r", "a", "w"] = "w",
        backend: typing.Literal["h5py", "adios2"] = "adios2",
        array_names: list[str] | None = None,
    ) -> None:
        if len(arrays) == 0:
            raise ValueError("No arrays to write to point cloud")
        for f in arrays:
            if not isinstance(f, np.ndarray):
                raise ValueError("All arrays must be of type numpy.ndarray")
        f0 = arrays[0]
        for f in arrays[1:]:
            if f.shape != f0.shape:
                raise ValueError("All arrays must have the same shape")

        self._data_arrays = arrays
        self._data = function_space_data
        # FIXME: Should we do some checks here on the data?

        if array_names is None:
            array_names = [f"array_{i}" for i in range(len(arrays))]
        self._data_names = array_names

        self.filename = filename
        self.filemode = filemode
        self.backend = backend
        super().__post_init__()

        self._init_backend()

    @property
    def data_names(self) -> list[str]:
        return self._data_names

    @property
    def data_arrays(self) -> list[npt.NDArray[np.floating]]:
        return self._data_arrays



import basix
import dolfinx
import ufl
import numpy.typing as npt
import numpy as np

__all__ = ["interpolate_function_onto_facet_dofs"]


def interpolate_function_onto_facet_dofs(
    Q: dolfinx.fem.FunctionSpace,
    expr: ufl.core.expr.Expr,
    facets: npt.NDArray[np.int32],
) -> dolfinx.fem.Function:
    """
    Create a function :math:`u_h\\in Q` such that :math:`u_h=\\text{expr}` for all dofs belonging
    to a subset of ``facets``. All other dofs are set to zero.

    Note:
        The resulting function  is only correct in the "normal" direction,
        i.e. :math:`u_{bc}\\cdot n = expr`, while the tangential component is uncontrolled.
        This makes it hard to visualize the function when outputting it to file, either
        through interpolation to an appropriate DG space, or to a point-cloud.

    Args:
        Q: The function space to create the function $u_h$ in.
        expr: The expression to evaluate.
        facets: The facets on which to evaluate the expression.
    """
    domain = Q.mesh
    Q_el = Q.element
    fdim = domain.topology.dim - 1
    domain.topology.create_connectivity(fdim, domain.topology.dim)

    interpolation_points = Q_el.basix_element.x

    c_el = domain.ufl_domain().ufl_coordinate_element()
    ref_top = c_el.reference_topology
    ref_geom = c_el.reference_geometry
    facet_types = set(basix.cell.subentity_types(domain.basix_cell())[fdim])
    assert len(facet_types) == 1, "All facets must have the same topology"

    # Pull back interpolation points from reference coordinate element to facet reference element
    facet_cmap = basix.ufl.element(
        "Lagrange",
        facet_types.pop(),
        c_el.degree,
        shape=(domain.geometry.dim,),
        dtype=np.float64,
    )
    facet_cel = dolfinx.cpp.fem.CoordinateElement_float64(facet_cmap.basix_element._e)
    reference_facet_points = None
    for i, points in enumerate(interpolation_points[fdim]):
        geom = ref_geom[ref_top[fdim][i]]

        ref_points = facet_cel.pull_back(points, geom)
        # Assert that interpolation points are all equal on all facets
        if reference_facet_points is None:
            reference_facet_points = ref_points
        else:
            assert np.allclose(reference_facet_points, ref_points)
    assert reference_facet_points is not None
    # Create expression for BC
    normal_expr = dolfinx.fem.Expression(expr, reference_facet_points)

    points_per_entity = [sum(ip.shape[0] for ip in ips) for ips in interpolation_points]
    offsets = np.zeros(domain.topology.dim + 2, dtype=np.int32)
    offsets[1:] = np.cumsum(points_per_entity[: domain.topology.dim + 1])
    values_per_entity = np.zeros(
        (offsets[-1], domain.geometry.dim), dtype=dolfinx.default_scalar_type
    )

    # Compute integration entities (cell, local_facet index) for all facets
    all_connected_cells = dolfinx.mesh.compute_incident_entities(
        domain.topology, facets, domain.topology.dim - 1, domain.topology.dim
    )
    values = np.zeros(len(all_connected_cells) * offsets[-1] * domain.geometry.dim)
    domain.topology.create_connectivity(domain.topology.dim, fdim)
    c_to_f = domain.topology.connectivity(domain.topology.dim, fdim)
    num_facets_on_process = (
        domain.topology.index_map(fdim).size_local + domain.topology.index_map(fdim).num_ghosts
    )
    is_marked = np.zeros(num_facets_on_process, dtype=np.int8)
    is_marked[facets] = 1
    for i, cell in enumerate(all_connected_cells):
        values_per_entity[:] = 0.0
        local_facets = c_to_f.links(cell)
        for j, lf in enumerate(local_facets):
            if not is_marked[lf]:
                continue
            insert_pos = offsets[fdim] + reference_facet_points.shape[0] * j
            # Backwards compatibility
            entity = np.array([[cell, j]], dtype=np.int32)
            try:
                normal_on_facet = normal_expr.eval(domain, entity)
            except (AttributeError, AssertionError):
                normal_on_facet = normal_expr.eval(domain, entity.flatten())
            # NOTE: evaluate within loop to avoid large memory requirements
            values_per_entity[insert_pos : insert_pos + reference_facet_points.shape[0]] = (
                normal_on_facet.reshape(-1, domain.geometry.dim)
            )
        values[
            i * offsets[-1] * domain.geometry.dim : (i + 1) * offsets[-1] * domain.geometry.dim
        ] = values_per_entity.reshape(-1)

    qh = dolfinx.fem.Function(Q)
    if hasattr(qh._cpp_object, "interpolate_f"):
        interpolate_func = qh._cpp_object.interpolate_f
    else:
        interpolate_func = qh._cpp_object.interpolate
    interpolate_func(values.reshape(-1, domain.geometry.dim).T.copy(), all_connected_cells)
    qh.x.scatter_forward()

    return qh



from mpi4py import MPI

from scifem import assemble_scalar, norm
import scifem.petsc
from dolfinx.mesh import create_unit_square, exterior_facet_indices
import dolfinx
import basix.ufl
from dolfinx.fem import (
    Function,
    functionspace,
    Constant,
    form,
    locate_dofs_topological,
    dirichletbc,
)
from dolfinx import default_scalar_type
import ufl
import pytest
import numpy as np


@pytest.mark.parametrize("gtype", [np.float64, np.float32])
def test_assemble_scalar_spatial(gtype):
    mesh = create_unit_square(MPI.COMM_WORLD, 3, 5, dtype=gtype)
    x = ufl.SpatialCoordinate(mesh)
    f = x[0] ** 2 * x[1] * ufl.dx
    tol = 50 * np.finfo(gtype).eps
    assert np.isclose(assemble_scalar(f), 1 / 3 * 1 / 2, atol=tol)


@pytest.mark.parametrize(
    "gtype, dtype",
    [
        [np.float64, np.float64],
        [np.float64, np.complex128],
        [np.float32, np.float32],
        [np.float32, np.complex64],
    ],
)
def test_assemble_scalar_constant(gtype, dtype):
    mesh = create_unit_square(MPI.COMM_WORLD, 3, 5, dtype=gtype)
    f = Constant(mesh, dtype(2.31))
    tol = 50 * np.finfo(gtype).eps
    assert np.isclose(assemble_scalar(f * ufl.dx), f.value, atol=tol)


@pytest.mark.parametrize(
    "gtype, dtype",
    [
        [np.float64, np.float64],
        [np.float64, np.complex128],
        [np.float32, np.float32],
        [np.float32, np.complex64],
    ],
)
def test_assemble_scalar_coefficient(gtype, dtype):
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5, dtype=gtype)
    V = functionspace(mesh, ("Lagrange", 3))
    u = Function(V, dtype=dtype)
    u.interpolate(lambda x: 3 * x[0] + 2 * x[1] ** 3)
    tol = 50 * np.finfo(gtype).eps
    assert np.isclose(assemble_scalar(u * ufl.dx), gtype(3 / 2 + 2 / 4), atol=tol)


def test_incompatible_coeff_function():
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5, dtype=np.float64)
    V = functionspace(mesh, ("Lagrange", 3))
    u = Function(V, dtype=np.float64)
    u.interpolate(lambda x: 3 * x[0] + 2 * x[1] ** 3)
    c = Constant(mesh, np.complex128(2.31))
    with pytest.raises(
        RuntimeError, match="All coefficients and constants must have the same data type"
    ):
        assemble_scalar(u * c * ufl.dx)


@pytest.mark.parametrize(
    "gtype, dtype",
    [
        [np.float64, np.float64],
        [np.float64, np.complex128],
        [np.float32, np.float32],
        [np.float32, np.complex64],
    ],
)
@pytest.mark.parametrize("norm_type", ["L2", "H1", "H10", "l2"])
def test_norm(norm_type, dtype, gtype):
    if norm_type == "l2":
        pytest.xfail("Unexpected norm type")

    mesh = create_unit_square(MPI.COMM_WORLD, 3, 5, dtype=gtype)
    V = functionspace(mesh, ("Lagrange", 1))
    u = Function(V, dtype=dtype)
    u.interpolate(lambda x: x[0] ** 2 + x[1] ** 2)

    x = ufl.SpatialCoordinate(mesh)
    expr = u - ufl.sin(ufl.pi * x[0])
    compiled_norm = norm(expr, norm_type)

    result = assemble_scalar(compiled_norm)

    if norm_type == "L2":
        ref_form = ufl.inner(expr, expr) * ufl.dx
    elif norm_type == "H1":
        ref_form = (
            ufl.inner(expr, expr) * ufl.dx + ufl.inner(ufl.grad(expr), ufl.grad(expr)) * ufl.dx
        )
    elif norm_type == "H10":
        ref_form = ufl.inner(ufl.grad(expr), ufl.grad(expr)) * ufl.dx

    reference = assemble_scalar(ref_form)
    tol = 50 * np.finfo(dtype).eps
    assert np.isclose(result, reference, atol=tol)


@pytest.mark.skipif(not dolfinx.has_petsc4py, reason="Requires DOLFINX with PETSc4py")
@pytest.mark.skipif(
    hasattr(dolfinx.fem.petsc, "create_matrix_nest"),
    reason="Requires latest version of DOLFINx PETSc API",
)
@pytest.mark.parametrize("kind", [None, "mpi", "nest"])
@pytest.mark.parametrize(
    "alpha", [dolfinx.default_scalar_type(3.0), dolfinx.default_scalar_type(-2.0)]
)
def test_lifting_helper(kind, alpha):
    from petsc4py import PETSc
    import dolfinx.fem.petsc

    mesh = create_unit_square(MPI.COMM_WORLD, 12, 15)
    el0 = basix.ufl.element("Lagrange", mesh.basix_cell(), 1)
    el1 = basix.ufl.element("Lagrange", mesh.basix_cell(), 2)

    if kind is None:
        el = basix.ufl.mixed_element([el0, el1])
        W = functionspace(mesh, el)
        V, _ = W.sub(0).collapse()
        Q, _ = W.sub(1).collapse()
    else:
        V = functionspace(mesh, el0)
        Q = functionspace(mesh, el1)
        W = ufl.MixedFunctionSpace(V, Q)

    k = Function(V)
    k.interpolate(lambda x: 3 + x[0] * x[1])
    u, p = ufl.TrialFunctions(W)
    v, q = ufl.TestFunctions(W)

    if kind is None:
        L = [
            ufl.inner(dolfinx.fem.Constant(mesh, dolfinx.default_scalar_type(0.0)), v) * ufl.dx,
            ufl.inner(dolfinx.fem.Constant(mesh, dolfinx.default_scalar_type(0.0)), q) * ufl.dx,
        ]
    else:
        L = [ufl.ZeroBaseForm((v,)), ufl.ZeroBaseForm((q,))]
    a = [
        [k * ufl.inner(u, v) * ufl.dx, None],
        [ufl.inner(u, q) * ufl.dx, k * ufl.inner(p, q) * ufl.dx],
    ]
    if kind is None:
        # Flatten a and L
        L = sum(L)
        a = [sum([aij if aij is not None else 0 for ai in a for aij in ai])]

    L_compiled = form(L)
    a_compiled = form(a)

    mesh.topology.create_connectivity(mesh.topology.dim - 1, mesh.topology.dim)
    bndry_facets = exterior_facet_indices(mesh.topology)

    bc_space = W.sub(1) if kind is None else Q

    bndry_dofs = locate_dofs_topological(bc_space, mesh.topology.dim - 1, bndry_facets)
    bcs = [dirichletbc(default_scalar_type(2.0), bndry_dofs, bc_space)]
    if kind is None:
        lifting_bcs = [bcs]
    else:
        lifting_bcs = bcs

    b = dolfinx.fem.petsc.create_vector(
        dolfinx.fem.forms.extract_function_spaces(L_compiled), kind=kind
    )
    scifem.petsc.zero_petsc_vector(b)
    scifem.petsc.apply_lifting_and_set_bc(b, a_compiled, lifting_bcs, alpha=alpha)

    # Create reference multiplication of alpha Ag
    if kind is None:
        a_compiled = a_compiled[0]
    A = dolfinx.fem.petsc.create_matrix(a_compiled, kind=kind)
    dolfinx.fem.petsc.assemble_matrix(A, a_compiled)
    A.assemble()

    g_ = dolfinx.fem.petsc.create_vector(
        dolfinx.fem.forms.extract_function_spaces(L_compiled), kind=kind
    )
    scifem.petsc.zero_petsc_vector(g_)

    if kind is not None:
        bcs0 = dolfinx.fem.bcs_by_block(dolfinx.fem.extract_function_spaces(a_compiled, 0), bcs)
    else:
        bcs0 = bcs
    dolfinx.fem.petsc.set_bc(g_, bcs0)

    b_ref = dolfinx.fem.petsc.create_vector(
        dolfinx.fem.forms.extract_function_spaces(L_compiled), kind=kind
    )
    A.mult(g_, b_ref)
    b_ref.scale(-alpha)
    dolfinx.fem.petsc.set_bc(b_ref, bcs0, alpha=alpha)
    scifem.petsc.ghost_update(b_ref, PETSc.InsertMode.INSERT_VALUES, PETSc.ScatterMode.FORWARD)
    with b.localForm() as b_local, b_ref.localForm() as b_ref_local:
        np.testing.assert_allclose(b_local.array, b_ref_local.array)



# # Point sources in DOLFINx
#
# Author: Jørgen S. Dokken
#
# SPDX-License-Identifier: MIT

# In this example, we will illustrate how to apply a point source to a Poisson problem.
#
# $$
# \begin{align}
# -\nabla^2 u &= f \quad \text{in } \Omega, \\
# f &= \gamma \delta(x - p) \quad \text{in } \Omega,
# \end{align}
# $$
#
# with homogeneous Dirichlet boundary conditions.
#
# Using integration by parts, we obtain the variational problem:
#
# Find $u_h\in V_0$ such that
#
# $$
# \begin{align}
# \int_\Omega \nabla u_h \cdot \nabla v \mathrm{d}x &= \int_\Omega \gamma\delta(x-p) v \mathrm{d}x
# = \gamma v(p) \quad \forall v \in V_0,
# \end{align}
# $$

# +
from mpi4py import MPI
from petsc4py import PETSc

import dolfinx
import ufl
import numpy as np
import scifem
# -

# We start by creating the mesh and function space.
# To illustrate the usage the point source function, we will
# consider a vector space with to components, and we will
# only apply the point source to the second component of our problem.

N = 40
domain = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, N, N)
domain.name = "mesh"
tdim = domain.topology.dim
domain.topology.create_connectivity(tdim-1, tdim)

# Create the two-component vector space.
# The solution in the first component will be 0.

value_shape = (2,)
V = dolfinx.fem.functionspace(domain, ("Lagrange", 2, value_shape))

# Create standard variational form

u = ufl.TrialFunction(V)
v = ufl.TestFunction(V)
a = ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx

# Create Dirichlet boundary conditions

boundary_facets = dolfinx.mesh.exterior_facet_indices(domain.topology)
dofs = dolfinx.fem.locate_dofs_topological(V, 1, boundary_facets)
u_bc = dolfinx.fem.Function(V)
u_bc.x.array[:] = 0
bc = dolfinx.fem.dirichletbc(u_bc, dofs)


# We are now ready to apply the point source.
# First, we create the right hand side vector for the problem, and initialize it as zero.

b = dolfinx.fem.Function(V)
b.x.array[:] = 0

# Secondly we define the point sources we want to apply.
#
# ```{warning}
# Note that if running in parallel, we only supply the points on a single process.
# The other process gets an empty array.
# ```

geom_dtype = domain.geometry.x.dtype
if domain.comm.rank == 0:
    points = np.array([[0.68, 0.362, 0],
                       [0.14, 0.213, 0]], dtype=geom_dtype)
else:
    points = np.zeros((0, 3), dtype=geom_dtype)

# Next, we create the point source object and apply it to the right hand side vector.

gamma = 1.
point_source = scifem.PointSource(V.sub(1), points, magnitude=gamma)
point_source.apply_to_vector(b)

# We can continue to solve our variational problem as usual

a_compiled = dolfinx.fem.form(a)
dolfinx.fem.petsc.apply_lifting(b.x.petsc_vec, [a_compiled], [[bc]])
b.x.scatter_reverse(dolfinx.la.InsertMode.add)
dolfinx.fem.petsc.set_bc(b.x.petsc_vec, [bc])
b.x.scatter_forward()

A = dolfinx.fem.petsc.assemble_matrix(a_compiled, bcs=[bc])
A.assemble()

# Set up a direct solver an solve the linear system

ksp = PETSc.KSP().create(domain.comm)
ksp.setOperators(A)
ksp.setType(PETSc.KSP.Type.PREONLY)
ksp.getPC().setType(PETSc.PC.Type.LU)
ksp.getPC().setFactorSolverType("mumps")


uh = dolfinx.fem.Function(V)
ksp.solve(b.x.petsc_vec, uh.x.petsc_vec)
uh.x.scatter_forward()

# We can now plot the solution

import pyvista

topology, cell_types, geometry = dolfinx.plot.vtk_mesh(V)
values = np.zeros((geometry.shape[0], 3), dtype=np.float64)
values[:, :len(uh)] = uh.x.array.real.reshape((geometry.shape[0], len(uh)))

# Create a point cloud of glyphs

function_grid = pyvista.UnstructuredGrid(topology, cell_types, geometry)
function_grid["u"] = values
glyphs = function_grid.glyph(orient="u", factor=0.2)

# Create a pyvista-grid for the mesh

domain.topology.create_connectivity(domain.topology.dim, domain.topology.dim)
grid = pyvista.UnstructuredGrid(*dolfinx.plot.vtk_mesh(domain, domain.topology.dim))

# Create plotter

plotter = pyvista.Plotter()
plotter.add_mesh(grid, style="wireframe", color="k")
plotter.add_mesh(glyphs)
plotter.view_xy()
if not pyvista.OFF_SCREEN:
    plotter.show()



import pytest

from petsc4py import PETSc
from mpi4py import MPI
import numpy as np
import dolfinx

from scifem import evaluate_function


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
def test_evaluate_scalar_function_2D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = 2.0
    nx = ny = 10

    mesh = dolfinx.mesh.create_rectangle(
        comm=comm,
        points=[np.array([0.0, 0.0]), np.array([Lx, Ly])],
        n=[nx, ny],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: x[0] + 2 * x[1]
    u.interpolate(f)

    points = np.array([[0.0, 0.0], [0.2, 0.2], [0.5, 0.5], [0.7, 0.2]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values[:, 0], exact)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
def test_evaluate_vector_function_2D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = 2.0
    nx = ny = 10

    mesh = dolfinx.mesh.create_rectangle(
        comm=comm,
        points=[np.array([0.0, 0.0]), np.array([Lx, Ly])],
        n=[nx, ny],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1, (2,)))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: (x[0], 2 * x[1])
    u.interpolate(f)

    points = np.array([[0.0, 0.0], [0.2, 0.2], [0.5, 0.5], [0.7, 0.2]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values, exact)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
def test_evaluate_scalar_function_3D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = Lz = 2.0
    nx = ny = nz = 10

    mesh = dolfinx.mesh.create_box(
        comm=comm,
        points=[np.array([0.0, 0.0, 0.0]), np.array([Lx, Ly, Lz])],
        n=[nx, ny, nz],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: x[0] + 2 * x[1]
    u.interpolate(f)

    points = np.array([[0.0, 0.0, 0.0], [0.2, 0.2, 0.3], [0.5, 0.5, 0.2], [0.7, 0.2, 0.5]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values[:, 0], exact)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
def test_evaluate_vector_function_3D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = Lz = 2.0
    nx = ny = nz = 10

    mesh = dolfinx.mesh.create_box(
        comm=comm,
        points=[np.array([0.0, 0.0, 0.0]), np.array([Lx, Ly, Lz])],
        n=[nx, ny, nz],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1, (3,)))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: (x[0], 2 * x[1], -x[2])
    u.interpolate(f)

    points = np.array([[0.0, 0.0, 0.0], [0.2, 0.2, 0.3], [0.5, 0.5, 0.2], [0.7, 0.2, 0.5]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values, exact)



import dolfinx
import basix.ufl
from pathlib import Path
import numpy as np
import numpy.typing as npt
import warnings


def apply_mri_transform(
    path: Path,
    coordinates: npt.NDArray[np.floating],
    vox2ras_transform: npt.NDArray[np.floating] | None = None,
    use_tkr: bool = False,
) -> npt.NDArray[np.inexact]:
    """Given a path to a set of MRI voxel data, return the data evaluated at the given coordinates.

    Args:
        path: Path to the MRI data, should be a file format supported by nibabel.
        coordinates: Coordinates to evaluate the MRI data at.
        vox2ras_transform: Optional transformation matrix to convert from voxel to ras coordinates.
            If None, use the transformation matrix from the given MRI data.
        use_tkr: If true, use the old freesurfer `tkregister`, see:
            https://www.mail-archive.com/freesurfer@nmr.mgh.harvard.edu/msg69541.html
            for more details. Else use the standard VOX2RAS transform (equivalent to attaching
            an affine map to a nibabel image.)

    Returns:
        The data evaluated at these points.
    """
    try:
        import nibabel
        import nibabel.affines as naff
    except ImportError:
        raise ImportError("This function requires the nibabel package to be installed")
    image = nibabel.load(path)
    data = image.get_fdata()
    # Check if the image is in the correct orientation
    orientation = nibabel.aff2axcodes(image.affine)
    if orientation != ("R", "A", "S"):
        warnings.warn(f"Image orientation is {orientation}, expected (R, A, S). ")

    # Define mgh_header depending on MRI data types (i.e. nifti or mgz)
    if isinstance(image, nibabel.freesurfer.mghformat.MGHImage):
        mgh_header = image.header
    elif isinstance(image, nibabel.nifti1.Nifti1Image):
        mgh = nibabel.MGHImage(image.dataobj, image.affine)
        mgh_header = mgh.header
    else:
        raise ValueError(
            f"Unsupported image type: {type(image)} - Supported types are mgz and nifti"
        )

    # This depends on how one uses FreeSurfer
    if vox2ras_transform is not None:
        vox2ras = vox2ras_transform
    else:
        if use_tkr:
            vox2ras = mgh_header.get_vox2ras_tkr()
        else:
            # VOX to ras explanation: https://surfer.nmr.mgh.harvard.edu/ftp/articles/vox2ras.pdf
            vox2ras = mgh_header.get_vox2ras()
    # Check shape
    if vox2ras.shape != (4, 4):
        raise ValueError(f"vox2ras transform must be a 4x4 matrix, got shape {vox2ras.shape}")

    ras2vox = np.linalg.inv(vox2ras)

    ijk_vectorized = naff.apply_affine(ras2vox, coordinates)
    # The file standard assumes that the voxel coordinates refer to the center of each voxel.
    # https://brainder.org/2012/09/23/the-nifti-file-format/
    ijk_rounded = np.rint(ijk_vectorized).astype("int")
    assert np.all(ijk_rounded >= 0)

    return data[ijk_rounded[:, 0], ijk_rounded[:, 1], ijk_rounded[:, 2]]


def read_mri_data_to_tag(
    mri_data_path: str | Path,
    mesh: dolfinx.mesh.Mesh,
    edim: int,
    entities: npt.NDArray[np.int32] | None = None,
    vox2ras_transform: npt.NDArray[np.floating] | None = None,
    use_tkr: bool = False,
) -> dolfinx.mesh.MeshTags:
    """Read in MRI data over a set of entities in the mesh and attach it to a mesh tag.

    Args:
        mri_data_path: Path to MRI data
        mesh: The mesh to attach the MRI data to.
        edim: Topological dimension of entities to attach the MRI data to.
        entities: List of entities (local to process) to evaluate the MRI data at. If None, all
            local entities are used.

    Returns:
        The mesh tag with the attached MRI data.
    """
    if entities is None:
        mesh.topology.create_entities(edim)
        entity_map = mesh.topology.index_map(edim)
        num_entities = entity_map.size_local + entity_map.num_ghosts
        entities = np.arange(num_entities, dtype=np.int32)

    midpoints = dolfinx.mesh.compute_midpoints(mesh, edim, entities)
    data = apply_mri_transform(Path(mri_data_path), midpoints, vox2ras_transform, use_tkr=use_tkr)
    et = dolfinx.mesh.meshtags(mesh, edim, entities, data.astype(np.int32))
    return et


def read_mri_data_to_function(
    mri_data_path: str | Path,
    mesh: dolfinx.mesh.Mesh,
    cells: npt.NDArray[np.int32] | None = None,
    vox2ras_transform: npt.NDArray[np.floating] | None = None,
    degree: int = 0,
    dtype: np.inexact = dolfinx.default_scalar_type,
    use_tkr: bool = False,
) -> dolfinx.fem.Function:
    """Read in MRI data over a set of cells in the mesh and attach it to an appropriate function.

    Args:
        mri_data_path: Path to MRI data
        mesh: The mesh to attach the MRI data to.
        cells: Subset of cells to evaluate the MRI data at. If None, all cells are used.
        vox2ras_transform: Optional transformation matrix to convert from voxel to ras coordinates.
            If None, use the transformation matrix from the given MRI data.
        degree: Degree of the (quadrature) function space to attach the MRI data to. Defaults to 0.
            If degree is 0, use a DG-0 space instead of a quadrature space, to simplify
            post-processing.
        dtype: Data type used for input data. Can be used for rounding data.
        use_tkr: If true, use the old freesurfer `tkregister`, see:
            https://www.mail-archive.com/freesurfer@nmr.mgh.harvard.edu/msg69541.html
            for more details. Else use the standard VOX2RAS transform (equivalent to attaching
            an affine map to a nibabel image.)


    Raises:
        ValueError: If degree is negative.

    Returns:
        The function with the attached MRI data.
    """
    if not degree >= 0:
        raise ValueError("Degree must be a non-negative integer")

    if degree == 0:
        el = basix.ufl.element("Lagrange", mesh.basix_cell(), 0, discontinuous=True)
    else:
        el = basix.ufl.quadrature_element(scheme="default", degree=degree, cell=mesh.basix_cell())

    if cells is None:
        cell_map = mesh.topology.index_map(mesh.topology.dim)
        num_cells = cell_map.size_local + cell_map.num_ghosts
        cells = np.arange(num_cells, dtype=np.int32)

    Vh = dolfinx.fem.functionspace(mesh, el)
    dof_positions = Vh.tabulate_dof_coordinates()
    dofmap = Vh.dofmap.list
    dofmap_pos = dofmap[cells].flatten()
    data = apply_mri_transform(
        Path(mri_data_path), dof_positions[dofmap_pos], vox2ras_transform, use_tkr=use_tkr
    )

    v = dolfinx.fem.Function(Vh)
    v.x.array[dofmap_pos] = data.astype(dtype)
    return v



from __future__ import annotations
from importlib.metadata import metadata

import dolfinx
import numpy as np
import numpy.typing as npt
from . import _scifem  # type: ignore
from .point_source import PointSource
from .assembly import assemble_scalar, norm
from .bcs import interpolate_function_onto_facet_dofs
from . import xdmf
from .solvers import BlockedNewtonSolver, NewtonSolver
from .spaces import create_real_functionspace, create_space_of_simple_functions
from .mesh import (
    compute_interface_data,
    compute_subdomain_exterior_facets,
    create_entity_markers,
    extract_submesh,
    find_interface,
    transfer_meshtags_to_submesh,
    reverse_mark_entities,
)
from .eval import evaluate_function
from .interpolation import interpolation_matrix, prepare_interpolation_data

meta = metadata("scifem")
__version__ = meta["Version"]
__author__ = meta.get("Author", "")
__license__ = meta.get("License", "MIT")
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]

__all__ = [
    "meta",
    "__version__",
    "__author__",
    "__license__",
    "__email__",
    "__program_name__",
    "PointSource",
    "assemble_scalar",
    "create_space_of_simple_functions",
    "compute_interface_data",
    "compute_subdomain_exterior_facets",
    "create_real_functionspace",
    "extract_submesh",
    "find_interface",
    "assemble_scalar",
    "PointSource",
    "xdmf",
    "vertex_to_dofmap",
    "dof_to_vertexmap",
    "create_entity_markers",
    "NewtonSolver",
    "BlockedNewtonSolver",
    "transfer_meshtags_to_submesh",
    "evaluate_function",
    "reverse_mark_entities",
    "norm",
    "interpolate_function_onto_facet_dofs",
    "interpolation_matrix",
    "prepare_interpolation_data",
]


if dolfinx.has_petsc4py:
    from .interpolation import petsc_interpolation_matrix  # type: ignore

    __all__ += ["petsc_interpolation_matrix"]


def vertex_to_dofmap(V: dolfinx.fem.FunctionSpace) -> npt.NDArray[np.int32]:
    """
    Create a map from the vertices (local to the process) to the correspondning degrees
    of freedom.

    Args:
        V: The function space

    Returns:
        An array mapping local vertex i to local degree of freedom

    Note:
        If using a blocked space this map is not unrolled for the DofMap block size.
    """
    return _scifem.vertex_to_dofmap(V.mesh._cpp_object.topology, V.dofmap._cpp_object)


def dof_to_vertexmap(V: dolfinx.fem.FunctionSpace) -> npt.NDArray[np.int32]:
    """
    Create a map from the degrees of freedom to the vertices of the mesh.
    As not every degree of freedom is associated with a vertex, every dof that is not
    associated with a vertex returns -1

    Args:
        V: The function space

    Returns:
        An array mapping local dof i to a local vertex
    """
    num_dofs_local = V.dofmap.index_map.size_local + V.dofmap.index_map.num_ghosts
    dof_to_vertex_map = np.full(num_dofs_local, -1, dtype=np.int32)
    v_to_d = _scifem.vertex_to_dofmap(V.mesh._cpp_object.topology, V.dofmap._cpp_object)
    dof_to_vertex_map[v_to_d] = np.arange(len(v_to_d), dtype=np.int32)
    return dof_to_vertex_map



# # Real function spaces
#
# Author: Jørgen S. Dokken
#
# License: MIT
#
# In this example we will show how to use the "real" function space to solve
# a singular Poisson problem.
#
# ## Mathematical formulation
#  The problem at hand is:
# Find $u \in H^1(\Omega)$ such that
#
# $$
# \begin{align}
# -\Delta u &= f \quad \text{in } \Omega, \\
# \frac{\partial u}{\partial n} &= g \quad \text{on } \partial \Omega, \\
# \int_\Omega u &= h.
# \end{align}
# $$
#
# ### Lagrange multiplier
# We start by considering the equivalent optimization problem:
# Find $u \in H^1(\Omega)$ such that
#
# $$
# \begin{align}
#   \mathop{\mathrm{arg\,min}}_{u \in H^1(\Omega)}~J(u) =
#   \mathop{\mathrm{arg\,min}}_{u \in H^1(\Omega)}~\frac{1}{2}\int_\Omega \vert \nabla u \cdot \nabla u \vert \mathrm{d}x
#   - \int_\Omega f u \mathrm{d}x - \int_{\partial \Omega} g u \mathrm{d}s,
# \end{align}
# $$
#
# such that
#
# $$
# \begin{align}
#   \int_\Omega u = h.
# \end{align}
# $$
#
# We introduce a Lagrange multiplier $\lambda$ to enforce the constraint:
#
# $$
# \begin{align}
#   \mathop{\mathrm{arg\,min}}_{u \in H^1(\Omega), \lambda\in \mathbb{R}}~\mathcal{L}(u, \lambda) =
#   \mathop{\mathrm{arg\,min}}_{u \in H^1(\Omega), \lambda\in \mathbb{R}} J(u) + \lambda (\int_\Omega u \mathrm{d}x-h).
# \end{align}
# $$
#
# We then compute the optimality conditions for the problem above
#
# $$
# \begin{align}
# \frac{\partial \mathcal{L}}{\partial u}[\delta u] &= \int_\Omega \nabla u \cdot \nabla \delta u \mathrm{d}x + \lambda\int \delta u \mathrm{d}x - \int_\Omega f \delta u ~\mathrm{d}x - \int_{\partial \Omega} g \delta u~\mathrm{d}s = 0, \\
# \frac{\partial \mathcal{L}}{\partial \lambda}[\delta \lambda] &=\delta \lambda (\int_\Omega u \mathrm{d}x -h)= 0.
# \end{align}
# $$
#
# We write the weak formulation:
#
# $$
# \begin{align}
# \int_\Omega \nabla u \cdot \nabla \delta u~\mathrm{d}x + \int_\Omega \lambda \delta u~\mathrm{d}x &= \int_\Omega f \delta u~\mathrm{d}x + \int_{\partial \Omega} g v \mathrm{d}s\\
# \int_\Omega u \delta \lambda  \mathrm{d}x &= h \delta \lambda .
# \end{align}
# $$
#
# where we have moved $\delta\lambda$ into the integral as it is a spatial constant.

# ## Implementation
# We start by import the necessary modules
# ```{admonition} Clickable functions/classes
# Note that for the modules imported in this example, you can click on the function/class
# name to be redirected to the corresponding documentation page.
# ```

# +
from packaging.version import Version
from mpi4py import MPI
from petsc4py import PETSc
import dolfinx.fem.petsc

import numpy as np
from scifem import create_real_functionspace, assemble_scalar
from scifem.petsc import apply_lifting_and_set_bc
import ufl
import pyvista
# -

# We start by creating the domain using {py:mod}`dolfinx` and derive the source terms
# $f$, $g$ and $h$ from our manufactured solution using {py:mod}`ufl`.
# For this example we will use the following exact solution
#
# $$
# \begin{align}
#   u_{exact}(x, y) = 0.3y^2 + \sin(2\pi x).
# \end{align}
# $$

# +
M = 20
mesh = dolfinx.mesh.create_unit_square(
    MPI.COMM_WORLD, M, M, dolfinx.mesh.CellType.triangle, dtype=np.float64
)
V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1))


def u_exact(x):
    return 0.3 * x[1] ** 2 + ufl.sin(2 * ufl.pi * x[0])


x = ufl.SpatialCoordinate(mesh)
n = ufl.FacetNormal(mesh)
g = ufl.dot(ufl.grad(u_exact(x)), n)
f = -ufl.div(ufl.grad(u_exact(x)))
h = assemble_scalar(u_exact(x) * ufl.dx)

# -

# ### Creating the real function space and mixed space
# We then create the Lagrange multiplier space using {py:func}`scifem.create_real_functionspace`.
# This creates a {py:class}`dolfinx.fem.FunctionSpace` with a single degree of freedom (constant) over the whole domain.

R = create_real_functionspace(mesh)

# Next, we can create a mixed-function space for our problem
#
# ```{admonition} Note on DOLFINx versions
# The API for creating blocked problems in DOLFINx has vastly improved over the last few versions.
# It is recommended to use DOLFINx v0.9.0 or later, where one can use {py:class}`ufl.MixedFunctionSpace`
# to create a mixed-space of `M` number of spaces. One can then use {py:func}`ufl.TrialFunctions`
# and {py:func}`ufl.TestFunctions` as one would do with a {py:func}`basix.ufl.mixed_element`.
# ```

if Version(dolfinx.__version__) == Version("0.8.0"):
    u = ufl.TrialFunction(V)
    lmbda = ufl.TrialFunction(R)
    du = ufl.TestFunction(V)
    dl = ufl.TestFunction(R)
elif Version(dolfinx.__version__) >= Version("0.9.0.0"):
    W = ufl.MixedFunctionSpace(V, R)
    u, lmbda = ufl.TrialFunctions(W)
    du, dl = ufl.TestFunctions(W)
else:
    raise RuntimeError("Unsupported version of dolfinx")

# ### Defining and assembling the variational problem
# We can now define the variational problem

# +
zero = dolfinx.fem.Constant(mesh, dolfinx.default_scalar_type(0.0))

a00 = ufl.inner(ufl.grad(u), ufl.grad(du)) * ufl.dx
a01 = ufl.inner(lmbda, du) * ufl.dx
a10 = ufl.inner(u, dl) * ufl.dx
L0 = ufl.inner(f, du) * ufl.dx + ufl.inner(g, du) * ufl.ds
L1 = ufl.inner(zero, dl) * ufl.dx

a = [[a00, a01], [a10, None]]
L = [L0, L1]
a_compiled = dolfinx.fem.form(a)
L_compiled = dolfinx.fem.form(L)
# -

# Note that we have defined the variational form in a block form, and
# that we have not included $h$ in the variational form. We will enforce this
# once we have assembled the right hand side vector.
# We can now assemble the matrix and vector usig {py:func}`dolfinx.fem.petsc.assemble_matrix`
# and {py:func}`dolfinx.fem.petsc.assemble_vector`.

if Version(dolfinx.__version__) < Version("0.10.0"):
    A = dolfinx.fem.petsc.assemble_matrix_block(a_compiled)
else:
    A = dolfinx.fem.petsc.assemble_matrix(a_compiled)
A.assemble()


# In DOLFINx>=v0.10.0, the `assemble_vector` function for blocked spaces has been rewritten to reflect how
# it works for standard assembly and `nest` assembly. This means that lifting is applied manually.
# In this case, with no Dirichlet BC, we could skip those steps.
# However, for clarity we include them here.

bcs = []


if Version(dolfinx.__version__) < Version("0.10.0"):
    b = dolfinx.fem.petsc.assemble_vector_block(L_compiled, a_compiled, bcs=bcs)
else:
    b = dolfinx.fem.petsc.assemble_vector(L_compiled, kind="mpi")
    apply_lifting_and_set_bc(b, a_compiled, bcs=bcs)

# Next, we modify the second part of the block to contain `h`
# We start by enforcing the multiplier constraint $h$ by modifying the right hand side vector.
# On the main branch, this is greatly simplified

# +
uh = dolfinx.fem.Function(V, name="u")

if Version(dolfinx.__version__) >= Version("0.10.0"):
    # We start by inserting the value in the real space
    rh = dolfinx.fem.Function(R)
    rh.x.array[0] = h
    # Next we need to add this value to the existing right hand side vector.
    # Therefore we create assign 0s to the primal space
    b_real_space = b.duplicate()
    uh.x.array[:] = 0
    # Transfer the data to `b_real_space`
    dolfinx.fem.petsc.assign([uh, rh], b_real_space)
    # And accumulate the values in the right hand side vector
    b.axpy(1, b_real_space)
    # We destroy the temporary work vector after usage
    b_real_space.destroy()
else:
    from dolfinx.cpp.la.petsc import scatter_local_vectors, get_local_vectors
    if Version(dolfinx.__version__) < Version("0.9.0"):
        maps = [(V.dofmap.index_map, V.dofmap.index_map_bs), (R.dofmap.index_map, R.dofmap.index_map_bs)]
    else:
        maps = [(Wi.dofmap.index_map, Wi.dofmap.index_map_bs) for Wi in W.ufl_sub_spaces()]

    b_local = get_local_vectors(b, maps)
    b_local[1][:] = h
    scatter_local_vectors(
            b,
            b_local,
            maps,
        )
    b.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)
# -

# ### Solving the linear system
# We can now solve the linear system using {py:mod}`petsc4py`.

# +
ksp = PETSc.KSP().create(mesh.comm)
ksp.setOperators(A)
ksp.setType("preonly")
pc = ksp.getPC()
pc.setType("lu")
pc.setFactorSolverType("mumps")

if Version(dolfinx.__version__) >= Version("0.10.0"):
    xh = b.duplicate()
else:
    xh = dolfinx.fem.petsc.create_vector_block(L_compiled)

ksp.solve(b, xh)
xh.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)
# -

# Finally, we extract the solution u from the blocked system and compute the error

uh = dolfinx.fem.Function(V, name="u")
if Version(dolfinx.__version__) >= Version("0.10.0"):
    dolfinx.fem.petsc.assign(xh, [uh, rh])
else:
    x_local = get_local_vectors(xh, maps)
    uh.x.array[: len(x_local[0])] = x_local[0]
    uh.x.scatter_forward()

# We destroy all PETSc objects

b.destroy()
xh.destroy()
A.destroy()
ksp.destroy()

# ### Post-processing
# Finally, we compare our approximate solution with the exact solution
# by computing the $L^2(\Omega)$ error.
# We use the convenience function {py:func}`scifem.assemble_scalar` to compute the error.


diff = uh - u_exact(x)
error = ufl.inner(diff, diff) * ufl.dx
print(f"L2 error: {np.sqrt(assemble_scalar(error)):.2e}")

# We can now plot the solution

# +
vtk_mesh = dolfinx.plot.vtk_mesh(V)

grid = pyvista.UnstructuredGrid(*vtk_mesh)
grid.point_data["u"] = uh.x.array.real

warped = grid.warp_by_scalar("u", factor=1)
plotter = pyvista.Plotter()
plotter.add_mesh(grid, style="wireframe")
plotter.add_mesh(warped)
if not pyvista.OFF_SCREEN:
    plotter.show()
# -



from mpi4py import MPI
import dolfinx
import numpy as np
import numpy.typing as npt


def evaluate_function(
    u: dolfinx.fem.Function, points: npt.ArrayLike, broadcast=True
) -> npt.NDArray[np.float64]:
    """Evaluate a function at a set of points.

    Args:
        u: The function to evaluate.
        points: The points to evaluate the function at.
        broadcast: If True, the values will be broadcasted to all processes.

            Note:
                Uses a global MPI call to broadcast values, thus this has to
                be called on all active processes synchronously.

            Note:
                If the function is discontinuous, different processes may return
                different values for the same point.
                In this case, the value returned is the maximum value across all processes.

    Returns:
        The values of the function evaluated at the points.


    """
    mesh = u.function_space.mesh
    u.x.scatter_forward()
    comm = mesh.comm
    points = np.array(points, dtype=np.float64)
    assert len(points.shape) == 2, (
        f"Expected points to have shape (num_points, dim), got {points.shape}"
    )
    num_points = points.shape[0]
    extra_dim = 3 - mesh.geometry.dim

    # Append zeros to points if the mesh is not 3D
    if extra_dim > 0:
        points = np.hstack((points, np.zeros((points.shape[0], extra_dim))))

    bb_tree = dolfinx.geometry.bb_tree(mesh, mesh.topology.dim)
    # Find cells whose bounding-box collide with the the points
    potential_colliding_cells = dolfinx.geometry.compute_collisions_points(bb_tree, points)
    # Choose one of the cells that contains the point
    adj = dolfinx.geometry.compute_colliding_cells(mesh, potential_colliding_cells, points)
    indices = np.flatnonzero(adj.offsets[1:] - adj.offsets[:-1])
    cells = adj.array[adj.offsets[indices]]
    points_on_proc = points[indices]

    values = u.eval(points_on_proc, cells)
    if broadcast:
        bs = u.function_space.dofmap.index_map_bs
        # Create array to store values and fill with -inf
        # to ensure that all points are included in the allreduce
        # with op=MPI.MAX
        u_out = np.ones((num_points, bs), dtype=np.float64) * -np.inf
        # Fill in values for points on this process
        u_out[indices, :] = values
        # Now loop over all processes and find the maximum value
        for i in range(num_points):
            if bs > 1:
                # If block size is larger than 1, loop over blocks
                for j in range(bs):
                    u_out[i, j] = comm.allreduce(u_out[i, j], op=MPI.MAX)
            else:
                u_out[i] = comm.allreduce(u_out[i], op=MPI.MAX)

        return u_out
    else:
        return values



from mpi4py import MPI
import dolfinx
import pytest
import ufl
import scifem
import numpy as np


def right_facets(x):
    return np.isclose(x[0], 1)


@pytest.mark.parametrize(
    "cell_type",
    [
        dolfinx.mesh.CellType.triangle,
        dolfinx.mesh.CellType.quadrilateral,
        dolfinx.mesh.CellType.hexahedron,
        dolfinx.mesh.CellType.tetrahedron,
    ],
)
def test_normal_enforcement(cell_type: dolfinx.mesh.CellType):
    tdim = dolfinx.mesh.cell_dim(cell_type)
    if tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 8, 10, 3)

    V = dolfinx.fem.functionspace(mesh, ("BDM", 1))

    n = ufl.FacetNormal(mesh)
    x = ufl.SpatialCoordinate(mesh)

    expr = x[1] * n

    facets = dolfinx.mesh.locate_entities_boundary(mesh, tdim - 1, right_facets)
    uh = scifem.interpolate_function_onto_facet_dofs(V, expr, facets)

    tag = 2
    facet_tags = dolfinx.mesh.meshtags(mesh, tdim - 1, facets, np.full_like(facets, tag))
    ds = ufl.ds(domain=mesh, subdomain_data=facet_tags, subdomain_id=tag)
    error = np.sqrt(
        scifem.assemble_scalar(ufl.inner(ufl.dot(uh - expr, n), ufl.dot(uh - expr, n)) * ds)
    )
    assert np.isclose(error, 0.0)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
def test_tangent_enforcement(cell_type: dolfinx.mesh.CellType):
    tdim = dolfinx.mesh.cell_dim(cell_type)
    if tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 8, 10, 3)
    else:
        raise ValueError(f"Unsupported {cell_type=}")

    V = dolfinx.fem.functionspace(mesh, ("N1curl", 2))

    n = ufl.FacetNormal(mesh)
    x = ufl.SpatialCoordinate(mesh)

    tangent = ufl.as_vector((-n[1], n[0]))

    expr = (0.1 * x[0] + x[1]) * tangent

    mesh.topology.create_connectivity(tdim - 1, tdim)
    facets = dolfinx.mesh.exterior_facet_indices(mesh.topology)
    uh = scifem.interpolate_function_onto_facet_dofs(V, expr, facets)

    error = np.sqrt(
        scifem.assemble_scalar(
            ufl.inner(ufl.dot(uh - expr, tangent), ufl.dot(uh - expr, tangent)) * ufl.ds
        )
    )
    assert np.isclose(error, 0.0)



from mpi4py import MPI
from packaging.version import Version
import dolfinx
import scifem.interpolation
import pytest
import ufl
import numpy as np
import basix.ufl


@pytest.mark.skipif(
    np.issubdtype(dolfinx.default_scalar_type, np.complexfloating), reason="No complex support"
)
@pytest.mark.parametrize(
    "cell_type",
    [
        dolfinx.mesh.CellType.triangle,
        dolfinx.mesh.CellType.quadrilateral,
        dolfinx.mesh.CellType.tetrahedron,
        dolfinx.mesh.CellType.hexahedron,
    ],
)
@pytest.mark.parametrize("use_petsc", [True, False])
@pytest.mark.parametrize("degree", [2, 4])
@pytest.mark.parametrize("out_family", ["Lagrange", "DG", "Quadrature"])
@pytest.mark.parametrize("value_shape", [(), (2,), (2, 3)])
def test_interpolation_matrix(use_petsc, cell_type, degree, out_family, value_shape):
    if use_petsc:
        pytest.importorskip("petsc4py")

    tdim = dolfinx.cpp.mesh.cell_dim(cell_type)
    if tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 4, 4, cell_type=cell_type)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 2, 2, cell_type=cell_type)
    else:
        raise ValueError("Unsupported cell type")

    V = dolfinx.fem.functionspace(mesh, ("DG", degree, value_shape))
    if out_family == "Quadrature":
        el = basix.ufl.quadrature_element(mesh.basix_cell(), degree=degree, value_shape=value_shape)
    else:
        el = (out_family, degree, value_shape)
    Q = dolfinx.fem.functionspace(mesh, el)

    def f(x):
        scalar_val = x[0] ** degree + x[1] if tdim == 2 else x[0] + x[1] + x[2] ** degree
        vs = int(np.prod(value_shape))
        f_rep = np.tile(scalar_val, vs).reshape(vs, -1)
        for i in range(vs):
            f_rep[i] += np.pi * (i + 1)
        return f_rep

    u = dolfinx.fem.Function(V)
    u.interpolate(f)

    q = dolfinx.fem.Function(Q)
    expr = ufl.TrialFunction(V)

    if use_petsc:
        A = scifem.interpolation.petsc_interpolation_matrix(expr, Q)
        A.mult(u.x.petsc_vec, q.x.petsc_vec)
        A.destroy()
    else:
        A = scifem.interpolation.interpolation_matrix(expr, Q)
        # Built in matrices has to use a special input vector, with additional ghosts.
        _x = dolfinx.la.vector(A.index_map(1), A.block_size[1])
        num_owned_dofs = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
        _x.array[:num_owned_dofs] = u.x.array[:num_owned_dofs]
        _x.scatter_forward()
        if not hasattr(dolfinx.la.MatrixCSR, "mult"):
            pytest.skip("MatrixCSR has no mult method")
        A.mult(_x, q.x)

    q.x.scatter_forward()

    q_ref = dolfinx.fem.Function(Q)
    if out_family == "Quadrature":
        try:
            ip = Q.element.interpolation_points()
        except TypeError:
            ip = Q.element.interpolation_points
        u_expr = dolfinx.fem.Expression(u, ip)
        q_ref.interpolate(u_expr)
    else:
        q_ref.interpolate(u)

    np.testing.assert_allclose(q.x.array, q_ref.x.array, rtol=1e-12, atol=1e-13)


@pytest.mark.skipif(
    np.issubdtype(dolfinx.default_scalar_type, np.complexfloating), reason="No complex support"
)
@pytest.mark.skipif(
    not hasattr(dolfinx.fem, "discrete_gradient"),
    reason="Cannot verify without discrete gradient from DOLFINx",
)
@pytest.mark.parametrize(
    "cell_type",
    [
        dolfinx.mesh.CellType.triangle,
        dolfinx.mesh.CellType.quadrilateral,
        dolfinx.mesh.CellType.tetrahedron,
        dolfinx.mesh.CellType.hexahedron,
    ],
)
@pytest.mark.parametrize("use_petsc", [True, False])
@pytest.mark.parametrize("degree", [1, 3, 5])
def test_discrete_gradient(degree, use_petsc, cell_type):
    if use_petsc:
        pytest.importorskip("petsc4py")

    tdim = dolfinx.cpp.mesh.cell_dim(cell_type)
    if tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 4, 4, cell_type=cell_type)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 2, 2, cell_type=cell_type)
    else:
        raise ValueError("Unsupported cell type")

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", degree))
    W = dolfinx.fem.functionspace(mesh, ("Nedelec 1st kind H(curl)", degree))

    u = dolfinx.fem.Function(V)
    u.interpolate(lambda x: x[0] ** degree + x[1])

    w = dolfinx.fem.Function(W)
    expr = ufl.grad(ufl.TrialFunction(V))

    G_ref = dolfinx.fem.discrete_gradient(V, W)

    # Built in matrices has to use a special input vector, with additional ghosts.
    try:
        _x = dolfinx.la.vector(G_ref.index_map(1), G_ref.block_size[1])
    except AttributeError:
        # Bug in DOLFINx 0.9.0
        _x = dolfinx.la.vector(G_ref.index_map(1), G_ref.bs[1])

    num_owned_dofs = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    _x.array[:num_owned_dofs] = u.x.array[:num_owned_dofs]
    _x.scatter_forward()

    if use_petsc:
        A = scifem.interpolation.petsc_interpolation_matrix(expr, W)
        A.mult(u.x.petsc_vec, w.x.petsc_vec)
        A.destroy()
    else:
        if not hasattr(dolfinx.la.MatrixCSR, "mult"):
            pytest.skip("MatrixCSR has no mult method")
        A = scifem.interpolation.interpolation_matrix(expr, W)
        A.mult(_x, w.x)
    w.x.scatter_forward()

    w_ref = dolfinx.fem.Function(W)
    if not hasattr(dolfinx.la.MatrixCSR, "mult"):
        # Fallback to PETSc discrete gradient on 0.9
        pytest.mark.skipif(not dolfinx.has_petsc4py, reason="Cannot verify without petsc4py")
        import dolfinx.fem.petsc as _petsc

        G_ref = _petsc.discrete_gradient(V, W)
        G_ref.assemble()
        G_ref.mult(u.x.petsc_vec, w_ref.x.petsc_vec)
    else:
        G_ref.mult(_x, w_ref.x)
    w_ref.x.scatter_forward()

    np.testing.assert_allclose(w.x.array, w_ref.x.array, rtol=1e-11, atol=1e-12)


@pytest.mark.skipif(
    np.issubdtype(dolfinx.default_scalar_type, np.complexfloating), reason="No complex support"
)
@pytest.mark.skipif(
    not hasattr(dolfinx.fem, "discrete_curl"),
    reason="Cannot verify without discrete curl from DOLFINx",
)
@pytest.mark.parametrize(
    "cell_type",
    [
        dolfinx.mesh.CellType.tetrahedron,
        dolfinx.mesh.CellType.hexahedron,
    ],
)
@pytest.mark.parametrize("use_petsc", [True, False])
@pytest.mark.parametrize("degree", [1, 2])
def test_discrete_curl(degree, use_petsc, cell_type):
    if use_petsc:
        pytest.importorskip("petsc4py")

    tdim = dolfinx.cpp.mesh.cell_dim(cell_type)
    if tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 4, 4, cell_type=cell_type)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 2, 2, cell_type=cell_type)
    else:
        raise ValueError("Unsupported cell type")

    V = dolfinx.fem.functionspace(mesh, ("Nedelec 2nd kind H(curl)", degree + 1))
    W = dolfinx.fem.functionspace(mesh, ("RT", degree))

    u = dolfinx.fem.Function(V)
    u.interpolate(lambda x: (x[0] ** degree, x[1] ** degree - 1, -x[2]))

    w = dolfinx.fem.Function(W)
    expr = ufl.curl(ufl.TrialFunction(V))

    G_ref = dolfinx.fem.discrete_curl(V, W)

    # Built in matrices has to use a special input vector, with additional ghosts.
    _x = dolfinx.la.vector(G_ref.index_map(1), G_ref.block_size[1])
    num_owned_dofs = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    _x.array[:num_owned_dofs] = u.x.array[:num_owned_dofs]
    _x.scatter_forward()

    if use_petsc:
        A = scifem.interpolation.petsc_interpolation_matrix(expr, W)
        A.mult(u.x.petsc_vec, w.x.petsc_vec)
        A.destroy()
    else:
        if not hasattr(dolfinx.la.MatrixCSR, "mult"):
            pytest.skip("MatrixCSR has no mult method")
        A = scifem.interpolation.interpolation_matrix(expr, W)
        A.mult(_x, w.x)
    w.x.scatter_forward()

    w_ref = dolfinx.fem.Function(W)
    G_ref.mult(_x, w_ref.x)
    w_ref.x.scatter_forward()

    np.testing.assert_allclose(w.x.array, w_ref.x.array, rtol=1e-10, atol=1e-11)


@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize("family", ["Lagrange"])
@pytest.mark.skipif(
    Version(dolfinx.__version__) < Version("0.10.0"), reason="Requires DOLFINx version >0.10.0"
)
def test_interpolate_to_interface_submesh(family, degree):
    # Create a unit square
    comm = MPI.COMM_WORLD
    domain = dolfinx.mesh.create_unit_square(
        comm, 48, 48, ghost_mode=dolfinx.mesh.GhostMode.shared_facet
    )

    # Split unit square in two subdomains
    cell_map = domain.topology.index_map(domain.topology.dim)
    num_cells_local = cell_map.size_local + cell_map.num_ghosts
    markers = np.full(num_cells_local, 1, dtype=np.int32)
    markers[
        dolfinx.mesh.locate_entities(domain, domain.topology.dim, lambda x: x[0] <= 0.5 + 1e-14)
    ] = 2
    ct = dolfinx.mesh.meshtags(
        domain, domain.topology.dim, np.arange(num_cells_local, dtype=np.int32), markers
    )

    # Create submesh for each subdomain
    omega_e, e_to_parent, _, _, _ = scifem.mesh.extract_submesh(domain, ct, (1,))
    omega_i, i_to_parent, _, _, _ = scifem.mesh.extract_submesh(domain, ct, (2,))

    # Compute submesh for the interface between omega_e and omega_i
    interface_facets = scifem.mesh.find_interface(ct, (1,), (2,))
    ft = dolfinx.mesh.meshtags(
        domain,
        domain.topology.dim - 1,
        interface_facets,
        np.full(interface_facets.shape, 1, dtype=np.int32),
    )

    gamma, gamma_to_parent, _, _, _ = scifem.mesh.extract_submesh(domain, ft, 1)

    num_facets_local = (
        gamma.topology.index_map(gamma.topology.dim).size_local
        + gamma.topology.index_map(gamma.topology.dim).num_ghosts
    )
    gamma_to_parent_map = gamma_to_parent.sub_topology_to_topology(
        np.arange(num_facets_local, dtype=np.int32), inverse=False
    )

    # Create functions on each subdomain
    def fe(x):
        return x[0] + x[1] ** degree

    def fi(x):
        return np.sin(x[0]) + np.cos(x[1])

    Ve = dolfinx.fem.functionspace(omega_e, (family, degree))
    ue = dolfinx.fem.Function(Ve)
    ue.interpolate(fe)
    ue.x.scatter_forward()
    Vi = dolfinx.fem.functionspace(omega_i, (family, degree))
    ui = dolfinx.fem.Function(Vi)
    ui.interpolate(fi)
    ui.x.scatter_forward()

    # Compute ordered integration entities on the interface
    interface_integration_entities = scifem.compute_interface_data(
        ct, facet_indices=gamma_to_parent_map, include_ghosts=True
    )
    mapped_entities = interface_integration_entities.copy()

    # For each submesh, get the relevant integration entities
    parent_to_e = e_to_parent.sub_topology_to_topology(
        np.arange(num_cells_local, dtype=np.int32), inverse=True
    )
    parent_to_i = i_to_parent.sub_topology_to_topology(
        np.arange(num_cells_local, dtype=np.int32), inverse=True
    )
    mapped_entities[:, 0] = parent_to_e[interface_integration_entities[:, 0]]
    mapped_entities[:, 2] = parent_to_i[interface_integration_entities[:, 2]]
    assert np.all(mapped_entities[:, 0] >= 0)
    assert np.all(mapped_entities[:, 2] >= 0)

    # Create two functions on the interface submesh
    Q = dolfinx.fem.functionspace(gamma, (family, degree))
    qe = dolfinx.fem.Function(Q, name="qe")
    qi = dolfinx.fem.Function(Q, name="qi")

    # Interpolate volume functions (on submesh) onto all cells of the interface submesh
    scifem.interpolation.interpolate_to_surface_submesh(
        ue, qe, np.arange(len(gamma_to_parent_map), dtype=np.int32), mapped_entities[:, :2]
    )
    qe.x.scatter_forward()
    scifem.interpolation.interpolate_to_surface_submesh(
        ui, qi, np.arange(len(gamma_to_parent_map), dtype=np.int32), mapped_entities[:, 2:]
    )
    qi.x.scatter_forward()

    # Compute the difference between the two interpolated functions
    I = dolfinx.fem.Function(Q, name="i")
    I.x.array[:] = qe.x.array - qi.x.array

    reference = dolfinx.fem.Function(Q)
    reference.interpolate(lambda x: fe(x) - fi(x))

    qe_ref = dolfinx.fem.Function(Q)
    qe_ref.interpolate(fe)
    qi_ref = dolfinx.fem.Function(Q)
    qi_ref.interpolate(fi)
    np.testing.assert_allclose(qe.x.array, qe_ref.x.array)
    np.testing.assert_allclose(qi.x.array, qi_ref.x.array)
    np.testing.assert_allclose(I.x.array, reference.x.array, rtol=1e-14, atol=1e-14)



from mpi4py import MPI
import numpy as np
import dolfinx
import scifem

import pytest


@pytest.mark.parametrize(
    "entities_list, set_values",
    [
        ([(1, lambda x: x[0] <= 1.0)], {1}),
        ([(2, lambda x: x[0] <= 1.0)], {2}),
        (
            [
                (1, lambda x: x[0] <= 1.0),
                (2, lambda x: x[0] >= 0.0),
            ],
            {2},
        ),
    ],
)
def test_create_celltags_celltags_all(entities_list, set_values):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 3, 3)
    cell_tag = scifem.create_entity_markers(mesh, mesh.topology.dim, entities_list=entities_list)

    im = mesh.topology.index_map(mesh.topology.dim)
    assert cell_tag.dim == mesh.topology.dim
    assert cell_tag.indices.shape[0] == im.size_local + im.num_ghosts
    assert cell_tag.values.shape[0] == im.size_local + im.num_ghosts
    assert cell_tag.values.dtype == np.int32
    assert set(cell_tag.values) == set_values


@pytest.mark.parametrize(
    "entities_list, set_values",
    [
        ([(1, lambda x: x[0] <= 1.0, False)], {1}),
        ([(2, lambda x: x[0] <= 1.0, False)], {2}),
        (
            [
                (1, lambda x: x[0] <= 1.0, False),
                (2, lambda x: x[0] >= 0.0, False),
            ],
            {2},
        ),
    ],
)
def test_create_facet_tags_all_on_boundary_False(entities_list, set_values):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 3, 3)
    facet_tag = scifem.create_entity_markers(
        mesh, mesh.topology.dim - 1, entities_list=entities_list
    )
    im = mesh.topology.index_map(mesh.topology.dim - 1)

    assert facet_tag.dim == mesh.topology.dim - 1
    assert facet_tag.indices.shape[0] == im.size_local + im.num_ghosts
    assert facet_tag.values.shape[0] == im.size_local + im.num_ghosts
    assert facet_tag.values.dtype == np.int32
    assert set(facet_tag.values) == set_values


@pytest.mark.parametrize(
    "entities_list, set_values",
    [
        ([(1, lambda x: x[0] <= 1.0, True)], {1}),
        ([(2, lambda x: x[0] <= 1.0, True)], {2}),
        (
            [
                (1, lambda x: x[0] <= 1.0, True),
                (2, lambda x: x[0] >= 0.0, True),
            ],
            {2},
        ),
    ],
)
def test_create_facet_tags_all_on_boundary_True(entities_list, set_values):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 3, 3)
    facet_tag = scifem.create_entity_markers(
        mesh, mesh.topology.dim - 1, entities_list=entities_list
    )

    mesh.topology.create_entities(1)
    mesh.topology.create_connectivity(mesh.topology.dim - 1, mesh.topology.dim)
    facet_indices = dolfinx.mesh.exterior_facet_indices(mesh.topology)
    assert facet_tag.dim == mesh.topology.dim - 1
    assert facet_tag.indices.shape[0] == len(facet_indices)
    assert facet_tag.values.shape[0] == len(facet_indices)
    assert facet_tag.values.dtype == np.int32
    assert set(facet_tag.values) == set_values


@pytest.mark.parametrize(
    "entities_list",
    [
        ([(1, lambda x: x[0] < 0.0)]),
        ([(1, lambda x: x[0] < 0.0)]),
        ([]),
    ],
)
def test_create_celltags_empty(entities_list):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 3, 3)
    cell_tag = scifem.create_entity_markers(mesh, mesh.topology.dim, entities_list=entities_list)

    assert cell_tag.dim == mesh.topology.dim
    assert cell_tag.indices.shape[0] == 0
    assert cell_tag.values.shape[0] == 0
    assert cell_tag.values.dtype == np.int32
    assert set(cell_tag.values) == set()


@pytest.mark.parametrize("edim", [0, 1, 2, 3])
def test_submesh_meshtags(edim):
    mesh = dolfinx.mesh.create_unit_cube(
        MPI.COMM_WORLD,
        3,
        4,
        7,
        cell_type=dolfinx.cpp.mesh.CellType.tetrahedron,
        ghost_mode=dolfinx.cpp.mesh.GhostMode.shared_facet,
    )

    mesh.topology.create_entities(edim)
    emap = mesh.topology.index_map(edim)

    # Put every second owned cell in submesh
    num_entities_local = emap.size_local
    subset_entities = np.arange(0, num_entities_local, 2, dtype=np.int32)
    # Include ghosts entities
    subset_cells = scifem.reverse_mark_entities(emap, subset_entities)

    submesh, entity_to_parent, vertex_to_parent, _ = dolfinx.mesh.create_submesh(
        mesh, edim, subset_cells
    )

    # Create meshtags on the parent mesh
    for i in range(edim + 1):
        mesh.topology.create_entities(i)
        parent_e_map = mesh.topology.index_map(i)
        num_parent_entities = parent_e_map.size_local + parent_e_map.num_ghosts
        values = parent_e_map.local_range[0] + np.arange(num_parent_entities, dtype=np.int32)
        entity_communicator = dolfinx.la.vector(parent_e_map, 1)
        entity_communicator.array[:] = values
        entity_communicator.scatter_forward()
        parent_tag = dolfinx.mesh.meshtags(
            mesh,
            i,
            np.arange(num_parent_entities, dtype=np.int32),
            entity_communicator.array.astype(np.int32),
        )
        sub_tag, sub_entity_to_parent = scifem.transfer_meshtags_to_submesh(
            parent_tag, submesh, vertex_to_parent, entity_to_parent
        )
        submesh.topology.create_connectivity(i, edim)
        midpoints = dolfinx.mesh.compute_midpoints(submesh, i, sub_tag.indices)
        mesh.topology.create_connectivity(i, mesh.topology.dim)
        parent_midpoints = dolfinx.mesh.compute_midpoints(mesh, i, parent_tag.indices)

        np.testing.assert_allclose(midpoints, parent_midpoints[sub_entity_to_parent])


@pytest.mark.parametrize("codim", [0, 1, 2])
@pytest.mark.parametrize("tdim", [1, 2, 3])
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.none, dolfinx.mesh.GhostMode.shared_facet]
)
def test_submesh_creator(codim, tdim, ghost_mode):
    edim = tdim - codim
    if edim < 0:
        pytest.xfail("Codim larger than tdim")

    if tdim == 1:
        mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 27, ghost_mode=ghost_mode)
    elif tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 18, ghost_mode=ghost_mode)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 7, 5, 8, ghost_mode=ghost_mode)
    else:
        raise ValueError("Invalid tdim")

    tol = 50 * np.finfo(mesh.geometry.x.dtype).eps

    def first_marker(x):
        return x[0] <= 0.5 + tol

    def second_marker(x):
        return x[tdim - 1] >= 0.6 - tol

    first_val = 2
    second_val = 3
    mesh.topology.create_entities(edim)
    emap = mesh.topology.index_map(edim)

    # Extract submesh no longer accumulates entities, so we mark the by hand
    indicator = dolfinx.la.vector(emap, 1, dtype=np.int32)
    indicator.array[:] = 0
    indicator.array[dolfinx.mesh.locate_entities(mesh, edim, first_marker)] = first_val
    indicator.array[dolfinx.mesh.locate_entities(mesh, edim, second_marker)] = second_val
    indicator.scatter_reverse(dolfinx.la.InsertMode.insert)
    indicator.scatter_forward()

    # Constructor we are testing
    etag = dolfinx.mesh.meshtags(
        mesh, edim, np.arange(emap.size_local + emap.num_ghosts, dtype=np.int32), indicator.array
    )
    submesh, cell_map, vertex_map, node_map, sub_etag = scifem.extract_submesh(
        mesh, etag, (first_val, second_val)
    )

    submap_array = scifem.mesh.get_entity_map(cell_map)
    parent_indices = submap_array[sub_etag.indices]
    np.testing.assert_allclose(sub_etag.values, indicator.array[parent_indices])

    # Create with standard constructor (reference)
    e_comm = dolfinx.la.vector(emap, 1)
    e_comm.array[:] = 0
    e_comm.array[dolfinx.mesh.locate_entities(mesh, edim, first_marker)] = 1
    e_comm.array[dolfinx.mesh.locate_entities(mesh, edim, second_marker)] = 1
    e_comm.scatter_reverse(dolfinx.la.InsertMode.add)
    e_comm.scatter_forward()
    sub_entities = np.flatnonzero(e_comm.array).astype(np.int32)
    ref_submesh, ref_cm, ref_vm, ref_nm = dolfinx.mesh.create_submesh(mesh, edim, sub_entities)

    submap_array = scifem.mesh.get_entity_map(cell_map)
    ref_submap_array = scifem.mesh.get_entity_map(ref_cm)
    np.testing.assert_allclose(submap_array, ref_submap_array)

    subvertexmap_array = scifem.mesh.get_entity_map(vertex_map)
    ref_subvertexmap_array = scifem.mesh.get_entity_map(ref_vm)
    np.testing.assert_allclose(subvertexmap_array, ref_subvertexmap_array)

    np.testing.assert_allclose(node_map, ref_nm)
    assert (
        submesh.topology.index_map(edim).size_local
        == ref_submesh.topology.index_map(edim).size_local
    )
    assert (
        submesh.topology.index_map(edim).size_global
        == ref_submesh.topology.index_map(edim).size_global
    )
    assert (
        submesh.topology.index_map(edim).num_ghosts
        == ref_submesh.topology.index_map(edim).num_ghosts
    )


@pytest.mark.parametrize("tdim", [1, 2, 3])
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.none, dolfinx.mesh.GhostMode.shared_facet]
)
@pytest.mark.parametrize("dtype", [np.float64, np.float32])
def test_find_interface(tdim, ghost_mode, dtype):
    comm = MPI.COMM_WORLD
    if tdim == 1:
        mesh = dolfinx.mesh.create_unit_interval(comm, 26, ghost_mode=ghost_mode, dtype=dtype)
    elif tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(comm, 10, 13, ghost_mode=ghost_mode, dtype=dtype)
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(comm, 6, 5, 8, ghost_mode=ghost_mode, dtype=dtype)
    else:
        raise ValueError("Invalid tdim")

    tol = 50 * np.finfo(dtype).eps

    def half_entities(x):
        return x[0] <= 0.5 + tol

    cell_map = mesh.topology.index_map(tdim)
    all_cells = np.arange(cell_map.size_local + cell_map.num_ghosts, dtype=np.int32)
    values = np.full_like(all_cells, 1, dtype=np.int32)
    values[dolfinx.mesh.locate_entities(mesh, tdim, half_entities)] = 2
    cell_tags = dolfinx.mesh.meshtags(mesh, tdim, all_cells, values)

    interface = scifem.find_interface(cell_tags, (1,), (2,))

    def ref_interface(x):
        return np.isclose(x[0], 0.5)

    ref_interface_facets = dolfinx.mesh.locate_entities(mesh, tdim - 1, ref_interface)
    mesh.topology.create_connectivity(tdim - 1, tdim)
    f_to_c = mesh.topology.connectivity(tdim - 1, tdim)
    ref_twosided_facets = []
    for f in ref_interface_facets:
        if len(f_to_c.links(f)) == 2:
            ref_twosided_facets.append(f)
    np.testing.assert_allclose(interface, ref_twosided_facets)


@pytest.mark.parametrize("dtype", [np.float64, np.float32])
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.none, dolfinx.mesh.GhostMode.shared_facet]
)
@pytest.mark.parametrize(
    "cell_type", [dolfinx.cpp.mesh.CellType.tetrahedron, dolfinx.cpp.mesh.CellType.hexahedron]
)
def test_exterior_boundary_subdomain(dtype, ghost_mode, cell_type):
    comm = MPI.COMM_WORLD
    mesh = dolfinx.mesh.create_unit_cube(
        comm, 13, 7, 11, cell_type=cell_type, ghost_mode=ghost_mode, dtype=dtype
    )

    def center(x):
        return (x[0] - 0.5) ** 2 + (x[1] - 0.5) ** 2 + (x[2] - 0.5) ** 2 <= 0.2**2

    tdim = mesh.topology.dim
    cell_map = mesh.topology.index_map(tdim)
    all_cells = np.arange(cell_map.size_local + cell_map.num_ghosts, dtype=np.int32)
    value_map = dolfinx.la.vector(cell_map, 1, dtype=np.int32)
    value_map.array[:] = 0
    value_map.array[dolfinx.mesh.locate_entities(mesh, tdim, center)] = 1
    value_map.scatter_reverse(dolfinx.la.InsertMode.add)
    value_map.scatter_forward()

    values = np.ones_like(all_cells, dtype=np.int32)
    values[value_map.array > 0] = 2

    cell_tags = dolfinx.mesh.meshtags(mesh, tdim, all_cells, values)

    def boundary_check(mesh, facets, cell_tags, values):
        is_in_subdomain = np.isin(cell_tags.values, np.asarray(values))
        mesh.topology.create_connectivity(mesh.topology.dim - 1, mesh.topology.dim)
        num_facets_local = mesh.topology.index_map(mesh.topology.dim - 1).size_local
        f_to_c = mesh.topology.connectivity(mesh.topology.dim - 1, mesh.topology.dim)
        cell_map = mesh.topology.index_map(mesh.topology.dim)
        owned_facets_in_subdomain = []
        owned_facets_on_outside = []
        ghosted_facets_in_subdomain = []
        ghosted_facets_on_outside = []

        facet_map = mesh.topology.index_map(mesh.topology.dim - 1)

        for facet in facets:
            cells = f_to_c.links(facet)
            if len(cells) != 1:
                assert np.sum(is_in_subdomain[cells]) == 1
            elif len(cells) == 1:
                if facet < num_facets_local:
                    # If connected to only one cell, we send it to the other process
                    # that has it to verify that it is a boundary facet
                    if is_in_subdomain[cells[0]]:
                        owned_facets_in_subdomain.append(facet)
                    else:
                        owned_facets_on_outside.append(facet)
                else:
                    if is_in_subdomain[cells[0]]:
                        ghosted_facets_in_subdomain.append(facet)
                    else:
                        ghosted_facets_on_outside.append(facet)
        # Receive facets from other process that is on subdomain on that process
        facet_vector = dolfinx.la.vector(
            mesh.topology.index_map(mesh.topology.dim - 1), 1, dtype=np.int32
        )
        facet_vector.array[:] = 0
        facet_vector.array[ghosted_facets_in_subdomain] = 1
        facet_vector.scatter_reverse(dolfinx.la.InsertMode.add)

        exterior_facets = dolfinx.mesh.exterior_facet_indices(mesh.topology)
        all_local_exterior_facets = scifem.reverse_mark_entities(facet_map, exterior_facets)

        facets_inside_subdomain = np.flatnonzero(facet_vector.array[:num_facets_local]).astype(
            np.int32
        )
        mesh.topology.create_connectivity(mesh.topology.dim, mesh.topology.dim - 1)
        for facet in facets_inside_subdomain:
            cells = f_to_c.links(facet)
            if facet in all_local_exterior_facets:
                assert is_in_subdomain[cells[0]]
                continue
            if len(cells) == 1:
                assert cells[0] < cell_map.size_local
                assert not is_in_subdomain[cells[0]]
            else:
                assert len(cells) == 2
                assert np.sum(is_in_subdomain[cells]) == 1

        facet_vector.array[:] = 0
        facet_vector.array[ghosted_facets_on_outside] = 1
        facet_vector.scatter_reverse(dolfinx.la.InsertMode.add)
        facets_outside_subdomain = np.flatnonzero(facet_vector.array[:num_facets_local]).astype(
            np.int32
        )
        for facet in facets_outside_subdomain:
            cells = f_to_c.links(facet)
            if facet in all_local_exterior_facets:
                assert is_in_subdomain[cells[0]]
                continue
            if len(cells) == 1:
                assert cells[0] < cell_map.size_local
                assert is_in_subdomain[cells[0]]
            else:
                assert len(cells) == 2
                assert np.sum(is_in_subdomain[cells]) == 1

        facet_vector.array[:] = 0
        facet_vector.array[owned_facets_in_subdomain] = 1
        facet_vector.scatter_forward()
        facets_inside_subdomain = np.flatnonzero(facet_vector.array[num_facets_local:]).astype(
            np.int32
        )
        for facet in facets_inside_subdomain:
            cells = f_to_c.links(num_facets_local + facet)
            if num_facets_local + facet in all_local_exterior_facets:
                assert is_in_subdomain[cells[0]]
                continue

            if len(cells) == 1:
                assert cells[0] <= cell_map.size_local
                assert not is_in_subdomain[cells[0]]
            else:
                assert len(cells) == 2
                assert np.sum(is_in_subdomain[cells]) == 1

        facet_vector.array[:] = 0
        facet_vector.array[owned_facets_on_outside] = 1
        facet_vector.scatter_forward()
        facets_outside_subdomain = np.flatnonzero(facet_vector.array[num_facets_local:]).astype(
            np.int32
        )
        for facet in facets_outside_subdomain:
            cells = f_to_c.links(num_facets_local + facet)
            if num_facets_local + facet in all_local_exterior_facets:
                assert is_in_subdomain[cells[0]]
                continue
            if len(cells) == 1:
                assert cells[0] <= cell_map.size_local
                assert is_in_subdomain[cells[0]]
            else:
                assert len(cells) == 2
                assert np.sum(is_in_subdomain[cells]) == 1

    ext_facets_1 = scifem.compute_subdomain_exterior_facets(mesh, cell_tags, (1,))
    boundary_check(mesh, ext_facets_1, cell_tags, (1,))

    # Exterior facets are only those at the interface
    ext_facets_2 = scifem.compute_subdomain_exterior_facets(mesh, cell_tags, (2,))
    boundary_check(mesh, ext_facets_2, cell_tags, (2,))

    # Exterior facets are only exterior
    ext_facets = scifem.compute_subdomain_exterior_facets(mesh, cell_tags, (1, 2))
    boundary_check(mesh, ext_facets, cell_tags, (1, 2))


@pytest.mark.parametrize(
    "cell_type",
    [
        dolfinx.mesh.CellType.tetrahedron,
        dolfinx.mesh.CellType.hexahedron,
        dolfinx.mesh.CellType.triangle,
        dolfinx.mesh.CellType.quadrilateral,
        dolfinx.mesh.CellType.interval,
    ],
)
@pytest.mark.parametrize("Nx", [6, 8, 11])
def test_compute_interface_data(cell_type: dolfinx.mesh.CellType, Nx: int):
    tdim = dolfinx.mesh.cell_dim(cell_type)
    if tdim == 1:
        mesh = dolfinx.mesh.create_unit_interval(
            MPI.COMM_WORLD, Nx, ghost_mode=dolfinx.mesh.GhostMode.shared_facet
        )
    elif tdim == 2:
        mesh = dolfinx.mesh.create_unit_square(
            MPI.COMM_WORLD,
            Nx,
            8,
            cell_type=cell_type,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
        )
    elif tdim == 3:
        mesh = dolfinx.mesh.create_unit_cube(
            MPI.COMM_WORLD,
            Nx,
            3,
            5,
            cell_type=cell_type,
            ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
        )
    else:
        raise ValueError(f"Invalid {cell_type=}")

    interface_loc = 3 / Nx

    def interface(x):
        return np.isclose(x[0], interface_loc)

    def left_domain(x, tol=100 * np.finfo(mesh.geometry.x.dtype).eps):
        return x[0] <= interface_loc + tol

    val_small = 3
    val_big = 8
    assert val_big > val_small
    tdim = mesh.topology.dim
    cell_map = mesh.topology.index_map(tdim)
    num_cells_local = cell_map.size_local + cell_map.num_ghosts
    values = np.full(num_cells_local, val_small, dtype=np.int32)
    values[dolfinx.mesh.locate_entities(mesh, tdim, left_domain)] = val_big
    cell_tags = dolfinx.mesh.meshtags(
        mesh, tdim, np.arange(num_cells_local, dtype=np.int32), values
    )

    mesh.topology.create_connectivity(tdim - 1, tdim)
    interface_facets = dolfinx.mesh.locate_entities(mesh, tdim - 1, interface)

    interface_data = scifem.compute_interface_data(cell_tags, interface_facets)

    assert np.isin(interface_data[:, 0], cell_tags.find(val_small)).all()
    assert np.isin(interface_data[:, 2], cell_tags.find(val_big)).all()



from mpi4py import MPI
import dolfinx
import basix
import numpy as np
from . import _scifem  # type: ignore
from collections.abc import Sequence
from packaging.version import Version


def create_real_functionspace(
    mesh: dolfinx.mesh.Mesh, value_shape: tuple[int, ...] = ()
) -> dolfinx.fem.FunctionSpace:
    """Create a real function space.

    Args:
        mesh: The mesh the real space is defined on.
        value_shape: The shape of the values in the real space.

    Returns:
        The real valued function space.
    Note:
        For scalar elements value shape is ``()``.

    """

    dtype = mesh.geometry.x.dtype
    ufl_e = basix.ufl.element(
        "P", mesh.basix_cell(), 0, dtype=dtype, discontinuous=True, shape=value_shape
    )

    if (dtype := mesh.geometry.x.dtype) == np.float64:
        cppV = _scifem.create_real_functionspace_float64(mesh._cpp_object, value_shape)
    elif dtype == np.float32:
        cppV = _scifem.create_real_functionspace_float32(mesh._cpp_object, value_shape)
    else:
        raise ValueError(f"Unsupported dtype: {dtype}")
    return dolfinx.fem.FunctionSpace(mesh, ufl_e, cppV)


def create_space_of_simple_functions(
    mesh: dolfinx.mesh.Mesh,
    cell_tag: dolfinx.mesh.MeshTags,
    tags: Sequence[int | np.int32 | np.int64],
    value_shape: tuple[()] | tuple[int | tuple[int]] | None = None,
) -> dolfinx.fem.FunctionSpace:
    """Create a space of simple functions.

    This is a space that represents piecewise constant functions of `N` patches,
    where `N` is the number of input `tags`.
    Each patch is defined by the cells in `cell_tag` which is marked with the
    corresponding tag value.

    Note:
        All cells are expected to have a tag in `tags`.

    Args:
        mesh: The mesh the function space is defined on.
        cell_tag: The mesh tags defining the different patches of cells.
        tags: The set of unique values within `cell_tag` that defines the different patches.
        value_shape: The shape of the values in the space.

    Returns:
        The space of simple functions.

    """
    value_shape = () if value_shape is None else value_shape
    if mesh.topology._cpp_object != cell_tag.topology:
        raise ValueError("Topology of cell tag is not the mesh topology")
    if cell_tag.dim != mesh.topology.dim:
        raise ValueError(
            f"The dimension of the input meshtag is {cell_tag.dim}, expected {mesh.topology.dim}"
        )

    cell_map = mesh.topology.index_map(mesh.topology.dim)

    # Sanity check cell tags
    if cell_map.size_local + cell_map.num_ghosts != len(cell_tag.indices):
        raise ValueError("Every cell in the cell tag must have a value assigned to it.")
    if not np.isin(cell_tag.values, tags).all():
        raise ValueError("All values in the cell tag must be in the input tags.")

    # Determine the owner of the material degrees of freedom.
    # It will be the process that owns the cell with global index 0
    is_owner = cell_map.local_range[0] == 0 and cell_map.size_local > 0

    # Determine number of dofs and ghosts on each process
    el = basix.ufl.element("P", mesh.basix_cell(), 0, shape=(), discontinuous=True)
    num_dofs = len(tags) * is_owner
    num_ghosts = len(tags) * (not is_owner)
    (_, ghost_owner_rank) = mesh.comm.allreduce((is_owner, mesh.comm.rank), op=MPI.MAXLOC)

    # Create index map describing the distribution of dofs across processes
    ghosts = np.arange(num_ghosts, dtype=np.int64)
    owners = np.full(num_ghosts, ghost_owner_rank, dtype=np.int32)
    if Version(dolfinx.__version__) > Version("0.9.0"):
        imap_kwargs = {"tag": 321}
    else:
        imap_kwargs = {}
    dof_imap = dolfinx.common.IndexMap(mesh.comm, num_dofs, ghosts, owners, **imap_kwargs)

    # Create element dof layout (1 dof per cell, based of the DG-0 element)
    value_size = int(np.prod(value_shape))
    index_map_bs = value_size
    bs = value_size

    e_layout = dolfinx.cpp.fem.ElementDofLayout(
        value_size, el.entity_dofs, el.entity_closure_dofs, [], []
    )

    # Create dofmap mapping local dofs to position in `tags` input.
    adj_flattened = np.zeros_like(cell_tag.values)
    for i, tag in enumerate(tags):
        adj_flattened[cell_tag.find(tag)] = i

    # Create dofmap
    dofmap_adj = dolfinx.cpp.graph.AdjacencyList_int32(
        adj_flattened, np.arange(len(cell_tag.values) + 1, dtype=np.int32)
    )
    cpp_dofmap = dolfinx.cpp.fem.DofMap(e_layout, dof_imap, index_map_bs, dofmap_adj, bs)

    # Create function space
    try:
        cpp_el = dolfinx.cpp.fem.FiniteElement_float64(
            el.basix_element._e, block_shape=value_shape, symmetric=False
        )
        cpp_space = dolfinx.cpp.fem.FunctionSpace_float64(mesh._cpp_object, cpp_el, cpp_dofmap)
    except TypeError:
        cpp_el = dolfinx.cpp.fem.FiniteElement_float64(
            el.basix_element._e, block_size=bs, symmetric=False
        )
        cpp_space = dolfinx.cpp.fem.FunctionSpace_float64(
            mesh._cpp_object, cpp_el, cpp_dofmap, value_shape=value_shape
        )

    _el = basix.ufl.element("P", mesh.basix_cell(), 0, shape=value_shape, discontinuous=True)
    V = dolfinx.fem.FunctionSpace(mesh, _el, cpp_space)
    return V



import pytest
import numpy as np
from mpi4py import MPI
import dolfinx
import ufl
from dolfinx.fem import Function, functionspace
from scifem.mesh import move


@pytest.fixture
def mesh_fixture():
    """
    Creates a fresh unit square mesh for each test.
    """
    comm = MPI.COMM_WORLD
    # Create a simple 2D unit square mesh
    mesh = dolfinx.mesh.create_unit_square(comm, 5, 5)
    return mesh


def test_move_with_dolfinx_function(mesh_fixture):
    """
    Test moving the mesh with a dolfinx.fem.Function.
    """
    mesh = mesh_fixture
    gdim = mesh.geometry.dim

    # 1. Create a displacement function in a Vector FunctionSpace
    V = functionspace(mesh, ("Lagrange", 2, (gdim,)))
    displacement = Function(V)

    # Define a simple constant displacement: u = (0.1, 0.2)
    disp_val = np.array([0.1, 0.2])

    # Using interpolate to set constant value for test clarity
    displacement.interpolate(lambda x: np.full((gdim, x.shape[1]), disp_val[:, None]))

    # 2. Store original coords to compare later
    original_coords = mesh.geometry.x.copy()

    # 3. Apply move
    move(mesh, displacement)

    # 4. Verify
    tol = 10 * np.finfo(mesh.geometry.x.dtype).eps

    expected_coords = original_coords[:, : mesh.geometry.dim] + disp_val[None, : mesh.geometry.dim]
    np.testing.assert_allclose(
        mesh.geometry.x[:, : mesh.geometry.dim],
        expected_coords,
        rtol=tol,
        atol=tol,
        err_msg="Mesh did not move correctly using dolfinx.fem.Function",
    )


def test_move_with_ufl_expression(mesh_fixture):
    """
    Test moving the mesh with a UFL Expression.
    """
    mesh = mesh_fixture

    # 1. Define UFL expression
    x = ufl.SpatialCoordinate(mesh)
    # Shift nodes radially: u = 0.1 * x
    ufl_expr = 0.1 * x

    # 2. Store original coords
    original_coords = mesh.geometry.x.copy()

    # 3. Apply move
    move(mesh, ufl_expr)

    # 4. Verify
    # The expected new position is x_new = x_old + 0.1 * x_old = 1.1 * x_old
    expected_coords = original_coords * 1.1
    tol = 10 * np.finfo(mesh.geometry.x.dtype).eps
    np.testing.assert_allclose(
        mesh.geometry.x,
        expected_coords,
        rtol=tol,
        atol=tol,
        err_msg="Mesh did not move correctly using UFL Expression",
    )


def test_move_with_python_callable(mesh_fixture):
    """
    Test moving the mesh with a generic Python callable.
    """
    mesh = mesh_fixture

    # 1. Define python callable
    # Input x has shape (3, N) or (gdim, N)
    # Output must match shape
    def shear_displacement(x):
        # Shear mapping: dx = 0.5 * y, dy = 0
        vals = np.zeros((mesh.geometry.dim, x.shape[1]), dtype=mesh.geometry.x.dtype)
        vals[0] = 0.5 * x[1]  # Displacement in x depends on y
        vals[1] = 0.0  # No displacement in y
        return vals

    # 2. Store original coords
    original_coords = mesh.geometry.x.copy()

    # 3. Apply move
    move(mesh, shear_displacement)

    # 4. Verify
    # Manually calculate expected shear
    expected_coords = original_coords.copy()
    expected_coords[:, 0] += 0.5 * original_coords[:, 1]

    tol = 10 * np.finfo(mesh.geometry.x.dtype).eps
    np.testing.assert_allclose(
        mesh.geometry.x,
        expected_coords,
        rtol=tol,
        atol=tol,
        err_msg="Mesh did not move correctly using Python callable",
    )


def test_move_accumulates_changes(mesh_fixture):
    """
    Ensure that calling move twice accumulates the displacement.
    """
    mesh = mesh_fixture
    # Move all coords by 0.1
    u = lambda x: np.full((mesh.geometry.dim, x.shape[1]), 0.1, dtype=mesh.geometry.x.dtype)
    original = mesh.geometry.x.copy()

    move(mesh, u)
    move(mesh, u)

    expected = original
    expected[:, : mesh.geometry.dim] += 0.2
    tol = 10 * np.finfo(mesh.geometry.x.dtype).eps
    np.testing.assert_allclose(
        mesh.geometry.x,
        expected,
        rtol=tol,
        atol=tol,
        err_msg="Multiple calls to move did not accumulate correctly",
    )



from mpi4py import MPI
import pytest
import dolfinx
import basix
import numpy as np
import ufl
from scifem import PointSource


def test_midpoint():
    """Check that adding point sources at all midpoints is equivalent of assembling
    the form ``v/abs(det(J))dx`` with a midpoint quadrature rule"""
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 2, 2)
    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1))
    tdim = mesh.topology.dim
    num_cells_local = mesh.topology.index_map(tdim).size_local
    midpoints = dolfinx.mesh.compute_midpoints(
        mesh, tdim, np.arange(num_cells_local, dtype=np.int32)
    )
    # Send all points from rank 0
    source_points = mesh.comm.gather(midpoints, root=0)
    if mesh.comm.rank == 0:
        source_points = np.vstack(source_points)
    else:
        source_points = np.empty((0, 3), dtype=midpoints.dtype)
    point_source = PointSource(V, source_points)
    np.testing.assert_allclose(point_source._cells, np.arange(num_cells_local, dtype=np.int32))
    np.testing.assert_allclose(point_source._points, midpoints)

    b = dolfinx.fem.Function(V)
    b.x.array[:] = 0
    point_source.apply_to_vector(b)
    b.x.scatter_reverse(dolfinx.la.InsertMode.add)
    b.x.scatter_forward()

    cell_geometry = basix.geometry(mesh.basix_cell())
    midpoint = (np.sum(cell_geometry, axis=0) / midpoints.shape[1]).reshape(
        -1, cell_geometry.shape[1]
    )

    b_ref = dolfinx.fem.Function(V)
    b_ref.x.array[:] = 0
    v = ufl.TestFunction(V)
    dx = ufl.Measure(
        "dx",
        domain=mesh,
        metadata={
            "quadrature_points": midpoint,
            "quadrature_weights": np.array([1.0]),
            "quadrature_rule": "custom",
        },
    )
    ref_L = dolfinx.fem.form(ufl.conj(v) / abs(ufl.JacobianDeterminant(mesh)) * dx)
    dolfinx.fem.assemble_vector(b_ref.x.array, ref_L)
    b_ref.x.scatter_reverse(dolfinx.la.InsertMode.add)
    b_ref.x.scatter_forward()

    np.testing.assert_allclose(b.x.array, b_ref.x.array)


def test_outside():
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 2, 2)
    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 3))
    if mesh.comm.rank == 0:
        point = np.array([[1.1, 1.8, 0]])
    else:
        point = np.empty((0, 3), dtype=mesh.geometry.x.dtype)

    if MPI.COMM_WORLD.rank == 0:
        with pytest.raises(ValueError, match="Point source is outside the mesh"):
            ps = PointSource(V, point)
    else:
        # Can only catch this on a single rank with pytest as it is
        # the only one that sends in a point
        ps = PointSource(V, point)

    # Check if perturbing the mesh gets the point within
    mesh.geometry.x[:, :] *= 2
    b = dolfinx.fem.Function(V)
    ps = PointSource(V, point)
    ps.apply_to_vector(b, recompute=False)
    b.x.scatter_reverse(dolfinx.la.InsertMode.add)
    b.x.scatter_forward()

    # Check sanity of values in b
    bbtree = dolfinx.geometry.bb_tree(mesh, mesh.topology.dim)
    cell_candidates = dolfinx.geometry.compute_collisions_points(bbtree, point)
    cells = dolfinx.geometry.compute_colliding_cells(mesh, cell_candidates, point)
    # Only use evaluate for points on current processor
    # BUG: In DOLFINx 0.8.0, links(0) yields a too long array
    if cells.offsets[-1] > 0:
        cell = cells.links(0)[0]
        geom_dofs = mesh.geometry.dofmap[cell]
        ref_x = mesh.geometry.cmap.pull_back(point.reshape(-1, 3), mesh.geometry.x[geom_dofs])
        ref_values = V.ufl_element().tabulate(0, ref_x).flatten()
        b_nonzero = np.flatnonzero(b.x.array)
        dofs = V.dofmap.cell_dofs(cell)
        assert len(b_nonzero) == len(dofs)
        assert len(np.intersect1d(b_nonzero, dofs)) == len(dofs)
        b_contributions = b.x.array[dofs]
        np.testing.assert_allclose(ref_values, b_contributions)

    # Scale down geometry and check that re-computing raises an error
    mesh.geometry.x[:, :] /= 2
    ps.apply_to_vector(b, recompute=False)

    if MPI.COMM_WORLD.rank == 0:
        with pytest.raises(ValueError, match="Point source is outside the mesh"):
            ps.apply_to_vector(b, recompute=True)
    else:
        ps.apply_to_vector(b, recompute=True)



# # Evaluating a function at a point
#
# Author: Henrik N.T. Finsberg
#
# SPDX-License-Identifier: MIT
#
# In this example we will show a how to to scifem for evaluating a function at a point.
# Note that the implementation is based on the approach outlined [here](https://jsdokken.com/FEniCS-workshop/src/deep_dive/expressions.html#evalation-at-a-point),
# and users are encouraged to read this for more details.

# Let us start by creating a rectangle mesh and a function space.

from mpi4py import MPI
import numpy as np
import dolfinx

from scifem import evaluate_function

comm = MPI.COMM_WORLD
Lx = Ly = 2.0
nx = ny = 10

mesh = dolfinx.mesh.create_rectangle(
    comm=comm,
    points=[np.array([0.0, 0.0]), np.array([Lx, Ly])],
    n=[nx, ny],
    cell_type=dolfinx.mesh.CellType.triangle
)

V = dolfinx.fem.functionspace(mesh, ("P", 1))
u = dolfinx.fem.Function(V)

# Now let us interpolate a function $f(x, y) = x + 2y$ into the function space.
# and use this as an example for evaluating the function at a set of points.

f = lambda x: x[0] + 2 * x[1]
u.interpolate(f)

# Let us pick a few points to evaluate the function at.

points = np.array([[0.0, 0.0], [0.2, 0.2], [0.5, 0.5], [0.7, 0.2]])

# The expected values of the function at the points are

exact = np.array(f(points.T)).T
print(exact)

# We can now evaluate the function at the points using the `evaluate_function` function.

u_values = evaluate_function(u, points)
print(u_values)
